
from escnn.gspaces import *
from escnn.nn import FieldType
from escnn.nn import GeometricTensor
from escnn.group import Representation
from escnn.group.representation import build_from_discrete_group_representation

from ..equivariant_module import EquivariantModule

import torch

from typing import List, Tuple, Any

import numpy as np
import math

__all__ = ["ConcatenatedNonLinearity"]


class ConcatenatedNonLinearity(EquivariantModule):
    
    def __init__(self, in_type: FieldType, function: str = 'c_relu'):
        r"""
        
        Concatenated non-linearities.
        For each input channel, the module applies the specified activation function both to its value and its opposite
        (the value multiplied by -1).
        The number of channels is, therefore, doubled.
        
        Notice that not all the representations support this kind of non-linearity. Indeed, only representations
        with the same pattern of permutation matrices and containing only values in :math:`\{0, 1, -1\}` support it.
        
        
        Args:
            in_type (FieldType): the input field type
            function (str): the identifier of the non-linearity. It is used to specify which function to apply.
                    By default (``'c_relu'``), ReLU is used.
            
        """
        
        assert isinstance(in_type.gspace, GSpace)
        
        for r in in_type.representations:
            assert 'concatenated' in r.supported_nonlinearities,\
                'Error! Representation "{}" does not support "concatenated" non-linearity'.format(r.name)

        super(ConcatenatedNonLinearity, self).__init__()

        self.space = in_type.gspace
        self.in_type = in_type
        
        # compute the output representation given the input one
        self.out_type = ConcatenatedNonLinearity._transform_fiber_representation(in_type)
        
        # retrieve the activation function to apply
        if function == 'c_relu':
            self._function = torch.relu
        elif function == 'c_sigmoid':
            self._function = torch.sigmoid
        elif function == 'c_tanh':
            self._function = torch.tanh
        else:
            raise ValueError('Function "{}" not recognized!'.format(function))
        
    def forward(self, input: GeometricTensor) -> GeometricTensor:
        
        assert input.type == self.in_type
        
        b, c = input.tensor.shape[:2]
        spatial_shape = input.tensor.shape[2:]
        
        # build the output tensor
        output = torch.empty(b, 2 * c, *spatial_shape, dtype=torch.float, device=input.tensor.device)
        
        # each channels is transformed to 2 channels:
        # first, apply the non-linearity to its value
        output[:, ::2, ...] = self._function(input.tensor)
        
        # then, apply the non-linearity to its values with the sign inverted
        output[:, 1::2, ...] = self._function(-1 * input.tensor)
        
        # wrap the result in a GeometricTensor
        return GeometricTensor(output, self.out_type, input.coords)

    def evaluate_output_shape(self, input_shape: Tuple[int, ...]) -> Tuple[int, ...]:

        assert len(input_shape) >= 2
        assert input_shape[1] == self.in_type.size

        b, c = input_shape[:2]
        spatial_shape = input_shape[2:]

        return (b, self.out_type.size, *spatial_shape)

    def check_equivariance(self, atol: float = 1e-6, rtol: float = 1e-5) -> List[Tuple[Any, float]]:
    
        c = self.in_type.size
    
        x = torch.randn(3, c, 10, 10)
    
        x = GeometricTensor(x, self.in_type)
    
        errors = []
    
        for el in self.space.testing_elements:
            out1 = self(x).transform_fibers(el)
            out2 = self(x.transform_fibers(el))
        
            errs = (out1.tensor - out2.tensor).detach().numpy()
            errs = np.abs(errs).reshape(-1)
            print(el, errs.max(), errs.mean(), errs.var())
        
            assert torch.allclose(out1.tensor, out2.tensor, atol=atol, rtol=rtol), \
                'The error found during equivariance check with element "{}" is too high: max = {}, mean = {} var ={}' \
                    .format(el, errs.max(), errs.mean(), errs.var())
        
            errors.append((el, errs.mean()))
    
        return errors

    @staticmethod
    def _transform_fiber_representation(in_type: FieldType) -> FieldType:
        r"""
        
        Compute the output representation from the input one after applying the concatenated non-linearity.
        
        Args:
            in_type (FieldType): the input field type

        Returns:
            (FieldType): the new output field type
            
        """
        transformed = {}
        
        # transform each different input Representation
        for repr in in_type._unique_representations:
            transformed[repr] = ConcatenatedNonLinearity._transform_representation(repr)
        
        new_representations = []
        
        # concatenate the new representations
        for repr in in_type.representations:
            new_representations.append(transformed[repr])
        
        return FieldType(in_type.gspace, new_representations)
        
    @staticmethod
    def _transform_representation(representation: Representation) -> Representation:
        r"""
        
        Transform an input :class:`~escnn.group.Representation` according to the concatenated non-linearity.
        
        The input representation needs to have the pattern of a permutation matrix, with values -1 or 1.
        
        The output representation has double the size of the input one and is built by substituting the ``1`` s with 2x2
        identity matrices and the ``-1`` s with 2x2 antidiagonal matrix containing ``1`` s.
        
        Args:
            representation (Representation): the input representation

        Returns:
            (Representation): the new output representation

        """
        group = representation.group
        
        assert not group.continuous
        
        # the name of the new representation
        name = "concatenated_{}".format(representation.name)
        
        if name in group.representations:
            # if the representation has already been built, return it
            r = group.representations[name]
        else:
            # otherwise, build the new representation
            
            s = representation.size
            
            rep = {}
            # build the representation for each element
            for element in group.elements:
                
                # retrieve the input representation of the current element
                r = representation(element)
                
                # build the matrix for the output representation of the current element
                rep[element] = np.zeros((2*s, 2*s))
                
                # check if the input matrix has the pattern of a permutation matrix
                e = [-1]*s
                for i in range(s):
                    for j in range(s):
                        if not math.isclose(r[i, j], 0, abs_tol=1e-9):
                            if e[i] < 0:
                                e[i] = j
                            else:
                                raise ValueError(
                                    '''Error! the representation should have the pattern of a permutation matrix
                                    but 2 values have been found in a row for element "{}"'''
                                    .format(element)
                                )
                if len(set(e)) != len(e):
                    raise ValueError(
                        '''Error! the representation should have the pattern of a permutation matrix
                        but 2 values have been found in a column for element "{}"'''
                        .format(element)
                    )
                
                # parse the input representation matrix and fill the output representation accordingly
                for i in range(s):
                    for j in range(s):
                        
                        if math.isclose(r[i, j], 1, abs_tol=1e-9):
                            # if the current cell contains 1, fill the output submatrix with the 2x2 identity
                            rep[element][2 * i:2 * i + 2, 2 * j:2 * j + 2] = np.eye(2)
                        elif math.isclose(r[i, j], -1, abs_tol=1e-9):
                            # if the current cell contains -1, fill the output submatrix with the 2x2 antidigonal matrix
                            rep[element][2 * i:2 * i + 2, 2 * j:2 * j + 2] = np.flipud(np.eye(2))
                        elif not math.isclose(r[i, j], 0, abs_tol=1e-9):
                            # otherwise the cell has to contain a 0
                            raise ValueError(
                                '''Error! The representation should be a signed permutation matrix and, therefore,
                                 contain only -1, 1 or 0 values but {} found in position({}, {}) for element "{}"'''
                                .format(r[i, j], i, j, element)
                            )
            
            # the resulting representation is a quotient repreentation and, therefore,
            # it also supports pointwise non-linearities
            nonlinearities = representation.supported_nonlinearities.union(['pointwise'])
            
            # build the output representation
            r = build_from_discrete_group_representation(rep, name, group, supported_nonlinearities=nonlinearities)
        
        return r






from .gspace import GSpace
from .r0 import *
from .r2 import *
from .r3 import *

# TODO move this sub-package to PyTorch part (e.g. under nn)

__all__ = [
    "GSpace",
    "GSpace0D",
    "GSpace2D",
    "GSpace3D",
    # R0
    "no_base_space",
    # R2
    "rot2dOnR2",
    "flipRot2dOnR2",
    "flip2dOnR2",
    "trivialOnR2",
    # R3
    "flipRot3dOnR3",
    "rot3dOnR3",
    "fullIcoOnR3",
    "icoOnR3",
    "fullOctaOnR3",
    "octaOnR3",
    "dihedralOnR3",
    "rot2dOnR3",
    "conicalOnR3",
    "fullCylindricalOnR3",
    "cylindricalOnR3",
    "mirOnR3",
    "invOnR3",
    "trivialOnR3"
]



from .equivariant_module import EquivariantModule

from .branching_module import BranchingModule
from .merge_module import MergeModule
from .multiple_module import MultipleModule

from .r2upsampling import R2Upsampling

from .conv import R3Conv
from .conv import R2Conv
from .conv import R2ConvTransposed
from .conv import R3ConvTransposed
from .conv import R3IcoConv
from .conv import R3IcoConvTransposed

try:
    from .pointconv import R2PointConv
    from .pointconv import R3PointConv
    _point_conv_modules = ['R2PointConv', 'R3PointConv']
except ImportError:
    _point_conv_modules = []


from .linear import Linear

from .nonlinearities import GatedNonLinearity1
from .nonlinearities import GatedNonLinearity2
from .nonlinearities import GatedNonLinearityUniform
from .nonlinearities import InducedGatedNonLinearity1
from .nonlinearities import NormNonLinearity
from .nonlinearities import InducedNormNonLinearity
from .nonlinearities import PointwiseNonLinearity
from .nonlinearities import ConcatenatedNonLinearity
from .nonlinearities import VectorFieldNonLinearity
from .nonlinearities import ReLU
from .nonlinearities import ELU
from .nonlinearities import FourierPointwise
from .nonlinearities import FourierELU
from .nonlinearities import QuotientFourierPointwise
from .nonlinearities import QuotientFourierELU
from .nonlinearities import TensorProductModule

from .reshuffle_module import ReshuffleModule

from .pooling import NormMaxPool
from .pooling import PointwiseMaxPool
from .pooling import PointwiseMaxPoolAntialiased
from .pooling import PointwiseAvgPool
from .pooling import PointwiseAvgPoolAntialiased
from .pooling import PointwiseAdaptiveAvgPool
from .pooling import PointwiseAdaptiveMaxPool
from .pooling import PointwiseAvgPool3D
from .pooling import PointwiseAvgPoolAntialiased3D

from .invariantmaps import GroupPooling
from .invariantmaps import MaxPoolChannels
from .invariantmaps import NormPool
from .invariantmaps import InducedNormPool

from .batchnormalization import InnerBatchNorm
from .batchnormalization import NormBatchNorm
from .batchnormalization import InducedNormBatchNorm
from .batchnormalization import GNormBatchNorm
from .batchnormalization import IIDBatchNorm1d, IIDBatchNorm2d, IIDBatchNorm3d

from .restriction_module import RestrictionModule
from .disentangle_module import DisentangleModule

from .dropout import FieldDropout
from .dropout import PointwiseDropout

from .sequential_module import SequentialModule
from .identity_module import IdentityModule

from .masking_module import MaskModule

from .harmonic_polynomial_r3 import HarmonicPolynomialR3


__all__ = [
    "EquivariantModule",
    "BranchingModule",
    "MergeModule",
    "MultipleModule",
    "Linear",
] + _point_conv_modules + [
    "R3Conv",
    "R2Conv",
    "R2ConvTransposed",
    "R3ConvTransposed",
    "R3IcoConv",
    "R3IcoConvTransposed",
    "R2Upsampling",
    "GatedNonLinearity1",
    "GatedNonLinearity2",
    "GatedNonLinearityUniform",
    "InducedGatedNonLinearity1",
    "NormNonLinearity",
    "InducedNormNonLinearity",
    "PointwiseNonLinearity",
    "ConcatenatedNonLinearity",
    "VectorFieldNonLinearity",
    "ReLU",
    "ELU",
    "FourierPointwise",
    "FourierELU",
    "QuotientFourierPointwise",
    "QuotientFourierELU",
    "TensorProductModule",
    "ReshuffleModule",
    "NormMaxPool",
    "PointwiseMaxPool",
    "PointwiseMaxPoolAntialiased",
    "PointwiseAvgPool",
    "PointwiseAvgPoolAntialiased",
    "PointwiseAvgPool3D",
    "PointwiseAvgPoolAntialiased3D",
    "PointwiseAdaptiveAvgPool",
    "PointwiseAdaptiveMaxPool",
    "GroupPooling",
    "MaxPoolChannels",
    "NormPool",
    "InducedNormPool",
    "InnerBatchNorm",
    "NormBatchNorm",
    "InducedNormBatchNorm",
    "GNormBatchNorm",
    "IIDBatchNorm1d",
    "IIDBatchNorm2d",
    "IIDBatchNorm3d",
    "RestrictionModule",
    "DisentangleModule",
    "FieldDropout",
    "PointwiseDropout",
    "SequentialModule",
    "IdentityModule",
    "MaskModule",
    "HarmonicPolynomialR3",
]




from .rd_convolution import _RdConv

from .r2convolution import R2Conv
from .r3convolution import R3Conv

from .r3_ico_convolution import R3IcoConv
from .r3_ico_transposed_convolution import R3IcoConvTransposed

from .r2_transposed_convolution import R2ConvTransposed
from .r3_transposed_convolution import R3ConvTransposed

__all__ = [
    "R3Conv",
    "R2Conv",
    "R2ConvTransposed",
    "R3ConvTransposed",
    "R3IcoConv",
    "R3IcoConvTransposed",
]




from escnn.group import Representation
from escnn.nn.modules import utils
from collections import defaultdict
from typing import List
import torch


def retrieve_indices(reprs: List[Representation]):
    fiber_position = 0
    _indices = defaultdict(list)
    _count = defaultdict(int)
    _contiguous = {}

    for repr in reprs:
        _indices[repr.name] += list(range(fiber_position, fiber_position + repr.size))
        fiber_position += repr.size
        _count[repr.name] += 1

    for name, indices in _indices.items():
        # _contiguous[o_name] = indices == list(range(indices[0], indices[0]+len(indices)))
        _contiguous[name] = utils.check_consecutive_numbers(indices)
        _indices[name] = torch.LongTensor(indices)

    return _count, _indices, _contiguous


########################################################################################################################


from .basismanager import BasisManager
from .basisexpansion_blocks import BlocksBasisExpansion
from .basisexpansion_singleblock import SingleBlockBasisExpansion
from .basissampler_blocks import BlocksBasisSampler
from .basissampler_singleblock import SingleBlockBasisSampler

__all__ = [
    "BasisManager",
    "BlocksBasisExpansion",
    "SingleBlockBasisExpansion",
    "BlocksBasisSampler",
    "SingleBlockBasisSampler",
]








from collections import defaultdict

import torch

from escnn.nn import GeometricTensor
from escnn.nn import FieldType
from .equivariant_module import EquivariantModule

from .reshuffle_module import ReshuffleModule

from escnn.gspaces import *

from typing import List, Tuple, Any, Dict

import numpy as np

__all__ = ["BranchingModule"]


class BranchingModule(EquivariantModule):
    
    def __init__(self,
                 in_type: FieldType,
                 labels: List[str],
                 reshuffle: int = 0
                 ):
    
        r"""
        
        Splits the input tensor in multiple branches identified by the input ``labels``.
        A label is associated to each field in the input class.
        During forward, fields are grouped by the labels and the input tensor is split accordingly,
        returning a dictionary mapping labels to tensors.
        
        If ``reshuffle`` is set to a positive integer, this module first builds a copy of the input tensor sorting the
        fields according to the value set:
        
        - 1: fields are sorted by their labels
        
        - 2: fields are sorted by their labels and, then, by their size
        
        - 3: fields are sorted by their labels, by their size and, then, by their type
        
        In this way, fields that need to be retrieved together are contiguous and it is possible to exploit slicing
        to split the tensor.
        By default, ``reshuffle = 0`` which means that no sorting is performed and, so, if input
        fields are not contiguous this layer will use indexing to retrieve sub-tensors.
        
        .. todo ::
            Technically this is not an EquivariantModule as the output is not a single tensor.
            Either fix EquivariantModule to support multiple inputs and outputs or set this as just subclass of
            torch.nn.Module.
        
        Args:
            in_type (FieldType): the input class
            labels (list): the list of labels to group the fields
            reshuffle (int, optional): set how to reshuffle the input fields before splitting the tensor.
                                       By default (``0``) no reshuffling is done.
            
        """
        
        assert isinstance(in_type.gspace, GSpace)
        
        assert 0 <= reshuffle < 3
        
        super(BranchingModule, self).__init__()
        
        self.space = in_type.gspace
        self.in_type = in_type
        self.reshuffle_layer = None
        
        self._labels = set(labels)
        
        total_fields = len(in_type.representations)
        
        assert total_fields == len(labels), \
            'Error! Number of labels ({}) does not match number of fields ({})'.format(len(labels), total_fields)
        
        # If the user required to sort the input representation build
        # a ReshuffleLayer to apply at the beginning of the forward
        if reshuffle > 0:
            
            # fields are sorted, in order of priority, by the non-linearity applied, their size and their name
            # according to the reshuffle set
            keys = []
            c = 0
            for l in labels:
                # build an array containing the sorting keys and the fields' positions
                keys.append((l, in_type.representations[c].size, in_type.representations[c].name, c))
                c += 1
            
            # sort the keys list to build the fields permutation
            keys = sorted(keys, key=lambda x: x[:reshuffle])
            permutation = [k[3] for k in keys]
            
            # if the fields were already sorted, it is useless to add the ReshuffleLayer
            if permutation != list(range(len(in_type.representations))):
                # build the ReshuffleLayer
                self.reshuffle_layer = ReshuffleModule(self.in_type, permutation)
                
                # add the reshuffle layer to the sub-modules
                self.add_module('reshuffle', self.reshuffle_layer)
                
                # set the input representation to consider for the non-linearities to the sorted one
                # (i.e. the output of the ReshuffleLayer)
                in_type = self.reshuffle_layer.out_type
                
                # permute the non-linearities list accordingly
                labels = [labels[p] for p in permutation]

        # for each label, build the representation of the sub-fiber on which it acts
        self.out_type = in_type.group_by_labels(labels)
        
        # check which non-linearity has all its fields consecutive
        self._contiguous = {}
        
        last_label = None
        for l in labels:
            if l != last_label:
                if not l in self._contiguous:
                    self._contiguous[l] = True
                else:
                    self._contiguous[l] = False
                
            last_label = l
        
        _input_indices = defaultdict(lambda: [])
        
        fields = defaultdict(lambda: [])
        
        # for each label, compute:
        #   - the set of indices on the fiber of its fields and
        #   - the the indices of the fields belonging to it
        c = 0
        last_position = 0
        for l in labels:
            # append the indices of the current field
            _input_indices[l] += list(range(last_position, last_position + in_type.representations[c].size))
            
            # append the index of the current field to the list of fields belonging to this label
            fields[l].append(c)
            
            # move on the fiber
            last_position += in_type.representations[c].size
            
            # move to the next field
            c += 1
        
        for l, contiguous in self._contiguous.items():
            if contiguous:
                # for labels with contiguous fields, only the first and the last indices are preserved
                _input_indices[l] = torch.LongTensor([min(_input_indices[l]), max(_input_indices[l])+1])
                
            else:
                # for the others, the indices list is trasformed into a PyTorch's Tensor
                _input_indices[l] = torch.LongTensor(_input_indices[l])

            # register the indices tensors as parameters of this module
            self.register_buffer('indices_{}'.format(l), _input_indices[l])
        
    def _retrieve_subfiber(self, input: GeometricTensor, l: str) -> GeometricTensor:
        r"""
        
        Return a new GeometricTensor containg the portion of memory of the input tensor corresponding to the fields
        the input non-linearity acts on. The method automatically deals with the continuity of these fields, using
        either indexing or slicing.
        
        The resulting tensor is returned wrapped in a GeometricTensor with the proper representation
        
        Args:
            input (GeometricTensor): the input tensor
            l (str): the label to consider

        Returns:
            (GeometricTensor): the sub-tensor containing the fields belonging to the input label
            
        """
        indices = getattr(self, f"indices_{l}")
        
        if self._contiguous[l]:
            # if the fields are contiguous, use slicing
            data = input.tensor[:, indices[0]:indices[1], ...]
        else:
            # otherwise, use indexing
            data = input.tensor[:, indices, ...]
        
        # wrap the result in a GeometricTensor
        return GeometricTensor(data, self.out_type[l], input.coords)
        
    def forward(self, input: GeometricTensor) -> Dict[str, GeometricTensor]:
        r"""
        
        Group input fields by their labels
        
        Args:
            input (GeometricTensor): the input feature map

        Returns:
            a dictionary mapping each label to a portion of the input tensor containing the
            fields assigned to that label
            
        """
        
        assert input.type == self.in_type
        
        # if reshuffling is required, apply the ReshuffleLayer
        if self.reshuffle_layer is not None:
            input = self.reshuffle_layer(input)
        
        output = {}
        
        # iterate through the labels
        for l in self._labels:
            # retrieve the corresponding sub-tensor
            output[l] = self._retrieve_subfiber(input, l)
            
        return output

    def evaluate_output_shape(self, input_shape: Tuple[int, ...]) -> Dict[Any, Tuple[int, ...]]:
        assert len(input_shape) == 4
        assert input_shape[1] == self.in_type.size
    
        b, c, hi, wi = input_shape
    
        return {l: (b, repr.size, hi, wi) for l, repr in self.out_type.items()}

    def check_equivariance(self, atol: float = 1e-6, rtol: float = 1e-5) -> List[Tuple[Any, float]]:
    
        c = self.in_type.size
    
        x = torch.randn(3, c, 1, 2)
    
        x = GeometricTensor(x, self.in_type)
    
        errors = []
    
        for el in self.space.testing_elements:
            
            outs2 = self(x.transform_fibers(el))
            
            outs1 = self(x)
            
            for l in self._labels:
                out1 = outs1[l].transform_fibers(el)
                out2 = outs2[l]
            
                errs = (out1.tensor - out2.tensor).detach().numpy()
                errs = np.abs(errs).reshape(-1)
                print(el, errs.max(), errs.mean(), errs.var())
            
                assert torch.allclose(out1.tensor, out2.tensor, atol=atol, rtol=rtol), \
                    'The error found during equivariance check with element "{}" is too high: max = {}, mean = {} var ={}' \
                        .format(el, errs.max(), errs.mean(), errs.var())
            
                errors.append((el, errs.mean()))
    
        return errors


from __future__ import annotations

import escnn.group
from escnn.group import Group, GroupElement

from .so3_utils import *
from .so3_utils import IDENTITY as IDENTITY_SO3
from .so3group import _clebsh_gordan_tensor_so3

from escnn.group import IrreducibleRepresentation
from escnn.group import Representation
from escnn.group import directsum
from escnn.group import change_basis
from escnn.group.irrep import restrict_irrep

from .utils import *

import numpy as np

from typing import Tuple, Callable, Iterable, List, Dict, Any, Union


__all__ = ["O3"]


class O3(Group):
    
    PARAM = PARAMETRIZATION

    PARAMETRIZATIONS = PARAMETRIZATIONS

    def __init__(self, maximum_frequency: int = 3):
        r"""
        Build an instance of the orthogonal group :math:`O(3)` which contains reflections and continuous 3D rotations.
        
        Subgroup structure:
        
        
        +-----------------------------------+-----------------------------------+-----------------------------------+---------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------+
        |    ``id[0]``                      |    ``id[1]``                      |    ``id[2]``                      |    subgroup                                                                                                                                                                                                   |
        +===================================+===================================+===================================+===============================================================================================================================================================================================================+
        |    `False`                        |        'so3'                      |                                   |   :math:`SO(3)` subgroup (equivalent to just `id = "so3"`)                                                                                                                                                    |
        +                                   +-----------------------------------+-----------------------------------+---------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------+
        |                                   |        'ico'                      |                                   |   Icosahedral :math:`I` subgroup                                                                                                                                                                              |
        +                                   +-----------------------------------+-----------------------------------+---------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------+
        |                                   |        'octa'                     |                                   |   Octahedral :math:`O` subgroup                                                                                                                                                                               |
        +                                   +-----------------------------------+-----------------------------------+---------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------+
        |                                   |        'tetra'                    |                                   |   Tetrahedral :math:`T` subgroup                                                                                                                                                                              |
        +                                   +-----------------------------------+-----------------------------------+---------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------+
        |                                   |        False                      |    -1                             |   :math:`SO(2)` subgroup of planar rotations around Z axis                                                                                                                                                    |
        +                                   +-----------------------------------+-----------------------------------+---------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------+
        |                                   |        False                      |     N                             |   :math:`C_N` of N discrete planar rotations around Z axis                                                                                                                                                    |
        +                                   +-----------------------------------+-----------------------------------+---------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------+
        |                                   |        True (or float)            |    -1                             |   *dihedral* :math:`O(2)` subgroup of planar rotations around Z axis and out-of-plane :math:`\pi` rotation around axis in the XY plane defined by ``id[1]/2`` (X axis by default)                             |
        +                                   +-----------------------------------+-----------------------------------+---------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------+
        |                                   |        True (or float)            |     N>1                           |   *dihedral* :math:`D_N` subgroup of N discrete planar rotations around Z axis and out-of-plane :math:`\pi` rotation around axis in the XY plane defined by ``id[1]/2`` (X axis by default)                   |
        +                                   +-----------------------------------+-----------------------------------+---------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------+
        |                                   |        True (or float)            |     1                             |   equivalent to ``(False, False, 2, adj)``                                                                                                                                                                    |
        +-----------------------------------+-----------------------------------+-----------------------------------+---------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------+
        |    'fulltetra'                    |                                   |                                   |   :math:`T_d \cong O` full isometry group tetrahedron                                                                                                                                                         |
        +-----------------------------------+-----------------------------------+-----------------------------------+---------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------+
        |    `True`                         |        'so3'                      |                                   |   :math:`O(3)` itself (equivalent to just `id = "o3"`)                                                                                                                                                        |
        +                                   +-----------------------------------+-----------------------------------+---------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------+
        |                                   |        'ico'                      |                                   |   Icosahedral :math:`I_h \cong I \times C_2` subgroup                                                                                                                                                         |
        +                                   +-----------------------------------+-----------------------------------+---------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------+
        |                                   |        'octa'                     |                                   |   Octahedral :math:`O_h \cong O \times C_2` subgroup                                                                                                                                                          |
        +                                   +-----------------------------------+-----------------------------------+---------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------+
        |                                   |        'tetra'                    |                                   |   Tetrahedral :math:`T_h \cong T \times C_2` subgroup                                                                                                                                                         |
        +                                   +-----------------------------------+-----------------------------------+---------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------+
        |                                   |        False                      |    -1                             |   :math:`SO(2) \times C_2` subgroup of planar rotations around Z axis and inversions                                                                                                                          |
        +                                   +-----------------------------------+-----------------------------------+---------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------+
        |                                   |        False                      |     N>1                           |   :math:`C_N \times C_2` of N discrete planar rotations around Z axis and inversions                                                                                                                          |
        +                                   +-----------------------------------+-----------------------------------+---------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------+
        |                                   |        False                      |     1                             |   :math:`C_2` subgroup inversions                                                                                                                                                                             |
        +                                   +-----------------------------------+-----------------------------------+---------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------+
        |                                   |        True (or float)            |    -1                             |   :math:`O(2) \times C_2` subgroup of planar rotations around Z axis, out-of-plane :math:`\pi` rotation around axis in the XY plane defined by ``id[1]/2`` (X axis by default) and inversions                 |
        +                                   +-----------------------------------+-----------------------------------+---------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------+
        |                                   |        True (or float)            |     N>1                           |   :math:`D_N \times C_2` subgroup of N discrete planar rotations around Z axis, out-of-plane :math:`\pi` rotation around axis in the XY plane defined by ``id[1]/2`` (X axis by default) and inversions       |
        +                                   +-----------------------------------+-----------------------------------+---------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------+
        |                                   |        True (or float)            |     1                             |   equivalent to ``(True, False, 2, adj)``                                                                                                                                                                     |
        +-----------------------------------+-----------------------------------+-----------------------------------+---------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------+
        |   'cone'                          |        -1                         |                                   |   *conic* subgroup :math:`O(2)`                                                                                                                                                                               |
        +                                   +-----------------------------------+-----------------------------------+---------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------+
        |                                   |         N>1                       |                                   |   *conic* subgroup of N discrete rotations :math:`D_N`                                                                                                                                                        |
        +                                   +-----------------------------------+-----------------------------------+---------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------+
        |                                   |         1                         |                                   |   subgroup :math:`C_2` of mirroring with respect to a plane passing through the X axis                                                                                                                        |
        +                                   +-----------------------------------+-----------------------------------+---------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------+
        |                                   |         float                     |    -1                             |   *conic* subgroup :math:`O(2)` with reflection along the ``id[1]/2`` axis in the XY plane. Equivalent to ('cone', -1, adj)                                                                                   |
        +                                   +-----------------------------------+-----------------------------------+---------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------+
        |                                   |         float                     |     N>1                           |   *conic* subgroup of N discrete rotations :math:`D_N` along the ``id[1]/2`` axis in the XY plane. Equivalent to ('cone', N, adj)                                                                             |
        +                                   +-----------------------------------+-----------------------------------+---------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------+
        |                                   |         float                     |     1                             |   subgroup :math:`C_2` of mirroring with respect to the plane passing through the ``id[1]/2`` axis in the XY plane. Equivalent to ('cone', 1, adj)                                                            |
        +-----------------------------------+-----------------------------------+-----------------------------------+---------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------+
        
        .. warning ::
            For some subgroups, ```id[1]``` expects either a boolean or a floating point value.
            In this case, ```id[1] = 0``` might be mistakenly interpreted as ```False``` rather than ```0.0```, so one
            should pay attention to use ```0.``` rather than ```0```.
        
        
        Args:
            maximum_frequency (int, optional): the maximum frequency to consider when building the irreps of the group
        
        """
        
        assert (isinstance(maximum_frequency, int) and maximum_frequency >= 0)
        
        super(O3, self).__init__("O(3)", True, False)

        self._maximum_frequency = maximum_frequency
        
        self._identity = self.element((0, IDENTITY_SO3))
        self._inversion = self.element((1, IDENTITY_SO3))
        
        self._build_representations()

    @property
    def generators(self) -> List[GroupElement]:
        raise ValueError(f'{self.name} is a continuous groups and '
                         f'some of its generators are infinitesimal. '
                         f'This is not currently supported')

    @property
    def identity(self) -> GroupElement:
        return self._identity

    @property
    def inversion(self) -> GroupElement:
        r"""
        The inversion element of :math:`\O3`.
        
        """
        return self._inversion

    @property
    def elements(self) -> List[GroupElement]:
        return None
     
    # @property
    # def elements_names(self) -> List[str]:
    #     return None

    @property
    def _keys(self) -> Dict[str, Any]:
        return dict()

    @property
    def subgroup_trivial_id(self):
        return (False, False, 1)

    @property
    def subgroup_self_id(self):
        return 'o3'

    ###########################################################################
    # METHODS DEFINING THE GROUP LAW AND THE OPERATIONS ON THE GROUP'S ELEMENTS
    ###########################################################################

    def _inverse(self, element: Tuple, param: str = PARAM) -> Tuple:
        r"""
        Return the inverse element of the input element.
        
        Args:
            element (tuple): a group element

        Returns:
            its inverse
            
        """
        
        return element[0], _invert(element[1], param)
        
    def _combine(self, e1: Tuple, e2: Tuple,
                param: str = PARAM,
                param1: str = None,
                param2: str = None
        ) -> Tuple:
        r"""
        Return the combination of the two input elements.
        Args:
            e1 (tuple): a group element
            e2 (tuple): another element

        Returns:
            their combination
        """
        return (e1[0] + e2[0]) % 2, _combine(e1[1], e2[1], param=param, param1=param1, param2=param2)

    def _equal(self, e1: Tuple, e2: Tuple,
              param: str = PARAM,
              param1: str = None,
              param2: str = None
        ) -> bool:
        r"""

        Check if the two input values corresponds to the same element.

        See :meth:`escnn.group.SO3.equal` for more details.

        Args:
            e1 (tuple): an element
            e2 (tuple): another element
            
        Returns:
            whether they are the same element

        """
        return (e1[0] == e2[0]) and _equal(e1[1], e2[1], param=param, param1=param1, param2=param2)

    def _hash_element(self, element: Tuple, param: str = PARAM):
        rot_hash = _hash(element[1], param)
        return hash(element[0]*100) + rot_hash

    def _repr_element(self, element: Tuple, param: str = PARAM):
    
        repr = _repr(element[1], param)
        repr = repr.split('\n')

        s = '+' if not element[0] else '-'
        repr[0] =    '     ' + repr[0]
        repr[1] = f'({s}), ' + repr[1]
        repr[2] =    '     ' + repr[2]
        return '\n'.join(repr)

    def _is_element(self, element: Tuple, param: str = PARAM, verbose: bool = False) -> bool:
        
        if not isinstance(element, Tuple) or len(element) != 2:
            if verbose:
                print(f'Expected a tuple of 2 elements, but {element} found')
            return False
        
        if element[0] not in {0, 1}:
            if verbose:
                print(f'The first value of an O(3) element needs to be an integer in {{0, 1}}, but {element[0]} found ')
            return False

        if not _check_param(element[1], param):
            if verbose:
                print(f'The second value of an O(3) element needs to be a rotation, but {element[1]} found ')
            return False
    
        return True

    def _change_param(self, element: Tuple, p_from, p_to) -> Tuple:
        assert p_from in self.PARAMETRIZATIONS
        assert p_to in self.PARAMETRIZATIONS
        
        if p_from == p_to:
            return element
        
        return element[0], _change_param(element[1], p_from, p_to)

    ###########################################################################
    
    def sample(self, param: str = PARAMETRIZATION) -> GroupElement:
        return self.element((np.random.randint(0, 2), _grid('rand', N=1, parametrization=param)[0]))

    def grid(self, type: str, *args, adj=IDENTITY, parametrization: str = PARAMETRIZATION, **kwargs) -> List[GroupElement]:
        r"""
        Method which builds different collections of elements of :math:`\O3`.

        Depending on the value of ``type``, the method accept a different set of parameters using the ``args`` and the
        ``kwargs`` arguments:

        * ``type = "rand"``. Generate ``N`` *random* uniformly distributed samples over the group. The method accepts an integer ``N`` and, optionally, a random seed ``seed`` (or an instance of :class:`numpy.random.RandomState`).
        
        Other options generate grids over :math:`\SO3` (see :meth:`escnn.group.SO3.grid`) and **then repeat it twice**,
        by combining each :math:`\SO3` element with both the identity and the *inversion* elements of :math:`\O3`.

        * ``type = "thomson"``.  Generate ``N`` samples distributed (approximately) uniformly over :math:`\SO3`

        * ``type = "ico", "cube", "tetra"``.  Generate respectively ``60``, ``24`` or ``12`` samples over :math:`\SO3`.

        * ``type = "hopf"``.  Generates *about* ``N`` points on :math:`\SO3`.

        * ``type = "fibonacci"``.  Generates *about* ``N`` points  on :math:`\SO3`.
        
        .. seealso ::
            :meth:`escnn.group.SO3.grid`

        Args:
            type (str): string identifying the type of samples
            *args: arguments specific for the type of samples chosen
            adj (GroupElement, optional): optionally, apply an adjoint transform to the sampled elements.
            parametrization (str, optional):
            **kwargs: arguments specific for the type of samples chosen

        Returns:
            a list of group elements

        """
        if type == 'rand':
            return [
                self.element(g, param=PARAMETRIZATION)
                for g in _random_samples(*args, **kwargs, parametrization=PARAMETRIZATION)
            ]
        else:
            so3_grid = _grid(type, *args, adj=adj, parametrization=parametrization, **kwargs)
            return [
                    self.element((0, g), param=parametrization) for g in so3_grid
            ] + [
                    self.element((1, g), param=parametrization) for g in so3_grid
            ]

    def grid_so3(self, type: str, *args, adj=IDENTITY, parametrization: str = PARAMETRIZATION, **kwargs) -> List[GroupElement]:
        r"""
        Method which builds different collections of elements of the :math:`\SO3` subgroup of :math:`\O3`.
        This method is equivalent to :meth:`escnn.group.SO3.grid`, but the :math:`\SO3` elements will be embedded in
        :math:`\O3`.

        Args:
            type (str): string identifying the type of samples
            *args: arguments specific for the type of samples chosen
            adj (GroupElement, optional): optionally, apply an adjoint transform to the sampled elements.
            parametrization (str, optional):
            **kwargs: arguments specific for the type of samples chosen

        Returns:
            a list of group elements

        """
        so3_grid = _grid(type, *args, adj=adj, parametrization=parametrization, **kwargs)
        return [self.element((0, g), param=parametrization) for g in so3_grid]

    def sphere_grid(self, type: str, *args, adj: GroupElement = None, **kwargs) -> List[GroupElement]:
        r"""

        Method which builds different collections of points over the sphere.

        Here, a sphere is interpreted as the quotient space :math:`\O3 / \O2` (where :math:`\O2` is the "cone" subgroup
        identified by the id ```('cone', -1)```).
        The list returned by this method contains instances of :class:`~escnn.group.GroupElement`.
        These are elements of :class:`~escnn.group.O3` and should be interpreted as *representatives* of cosets in
        :math:`\O3 / \O2`.

        Depending on the value of ``type``, the method accept a different set of parameters using the ``args`` and the
        ``kwargs`` arguments:

        * ``type = "rand"``. Generate ``N`` *random* uniformly distributed samples over the sphere. The method accepts an integer ``N`` and, optionally, a random seed ``seed`` (or an instance of :class:`numpy.random.RandomState`).

        * ``type = "thomson"``.  Generate ``N`` samples distributed (approximately) uniformly over the sphere. The samples are obtained with a numerical method which tries to minimize a potential energy depending on the relative distance between the points. The first call of this method for a particular value of ``N`` might be slow due to this numerical method. However, the resulting set of points is cached on disk such that following calls will return the same result instantly.

        * ``type = "thomson_cube"``.  Generate ``24*N`` samples with perfect *cubic* (*octahedral*) symmetry, distributed (approximately) uniformly over the shere. The samples are obtained with a numerical method which tries to minimize a potential energy depending on the relative distance between the points. The first call of this method for a particular value of ``N`` might be slow due to this numerical method. However, the resulting set of points is cached on disk such that following calls will return the same result instantly.

        * ``type = "ico", "cube", "tetra"``.  Return the orbit of the point :math:`(0, 0, 1)` under the action of the Icosahedral, Octahedral or Tetrahedral group. The resulting samples correspond to the corners, edges or faces of a Platonic solid.

        .. todo ::
            add more types to support vertices, edges and faces for all solids.
            Currently, ``"ico"`` returns the centers of the 30 edges while ``"cube"`` returns the centers of its 6 faces

        .. warning::
            The naming convention might change in future to support faces/edges/vertices of the solids

        .. note ::
            ``type = "tetra"`` does not have inversion symmetry but mirror symmetries (see Symmetries of the Tetrahedron)

        * ``type = "healpix"``.  Generates *about* ``N`` points using the HEAPlix grid on the sphere.

        * ``type = "fibonacci"``.  Generates ``N`` points using the Fibonacci grid on the sphere.

        * ``type = "longlat"``.  Generates an ``N`` by ``M`` grid over the sphere in the longitude (``N``) - latitude (``M``) coordinates.


        Args:
            type (str): string identifying the type of samples
            *args: arguments specific for the type of samples chosen
            adj (GroupElement, optional): optionally, apply an adjoint transform to the sampled elements.
            **kwargs: arguments specific for the type of samples chosen

        Returns:
            a list of group elements

        """

        if adj is None:
            adj = self.identity
        adj = self.standard_representation()(adj)

        points = _sphere_grid(type, *args, adj=adj, **kwargs)

        S = points.shape[0]
        x, y, z = points.T

        theta = np.arccos(np.clip(z, -1., 1.))
        phi = np.arctan2(y, x)

        return [self.element((0, (0, theta[s], phi[s])), 'zyz') for s in range(S)]

    def testing_elements(self, n=3) -> Iterable[GroupElement]:
        r"""
        A finite number of group elements to use for testing.
        """
        return self.grid('ico')

        so3_samples = np.empty((n**3, 3))
        i = 0
        for a in range(n):
            alpha = 2 * np.pi * a / n
            for b in range(n):
                # beta = np.pi * (2 * b + 1) / (2*n)
                # beta = np.pi * (b + 1) / (n+2)
                # sample around b = 0  and b = np.pi to check singularities
                beta = np.pi * b / n
                for c in range(n):
                    gamma = 2 * np.pi * c / n
                    so3_samples[i, :] = alpha, beta, gamma
                    i+=1

        so3_samples = _change_param(so3_samples, p_from='ZYZ', p_to=self.PARAM)

        samples = [self.element((0, g), self.PARAM) for g in so3_samples]
        samples += [self.element((1, g), self.PARAM) for g in so3_samples]

        return samples

    def __eq__(self, other):
        if not isinstance(other, O3):
            return False
        else:
            return self.name == other.name # and self._maximum_frequency == other._maximum_frequency

    def _process_subgroup_id(self, id):
        
        # (True, None)   -> Inversion C_2 subgroup
        # (False, x)     -> SO(3) subgroup (x)
        # (True, x)      -> SO(3) subgroup (x)  X Inversion
        #   includes also Cylinder as O(2) X Inversion
        # ('fulltetra')  -> full tethraedral is a semidirect prod however; it is however isomorphic to Octa group!!!
        # ('cone', N)    -> conical subgroup, O(2) or D_N depending on N

        if not isinstance(id, tuple):
            id = (id,)

        if id[0] == 'so3':
            id = False, 'so3', *id[1:]

        if id[0] == 'o3':
            id = True, 'so3', *id[1:]
            
        if not isinstance(id[-1], GroupElement):
            id = (*id, self.identity)
    
        assert id[-1].group == self
        
        if len(id) == 4 and isinstance(id[0], bool):
            assert isinstance(id[1], float) or isinstance(id[1], bool)

        if len(id) == 4 and (
                isinstance(id[0], bool) and isinstance(id[1], float)
        ):
            # O(2) (or subgroup) subgroup of SO(3)
            # if id[1]  is bool, it specifies SO(2) vs O(2) subgroup
            # if id[1] is float, it assumes O(2) and indicates the flip axis of O(2)
            # in this last case, we convert it to the boolean convention and include the flip axis inside the adjoing

            # The factor of 2 in id[1] comes from the fact that a flip around the axis theta, as an element of O(2), is
            # the combination of a reflection along the X axis (theta=0) and a rotation by 2*theta.
            # The values id[1] = 2*theta represents the SO(2) component of the flip.

            flip_axis = id[1] / 2.
            flip = np.asarray([0., 0., np.sin(flip_axis / 2), np.cos(flip_axis / 2)])
            flip = self.element((0, flip), 'Q')
            adjoint = (~flip) @ id[-1]
            id = (id[0], True, id[2], adjoint)

        if len(id) == 4 and (
                isinstance(id[0], bool) and id[1] and id[2] == 1
        ):
            # flip subgroup of the O(2) subgroup of SO(3) subgroup
            # this is equivalent to the C_2 subgroup of 180 deg rotations out of the plane
            change_axis = np.asarray([0., np.sin(-np.pi / 4), 0., np.cos(-np.pi / 4)])
            adj = self.element((0, change_axis), 'Q') @ id[3]
            id = (id[0], False, 2, adj)

        if id[0] not in ['fulltetra', 'cone']:
            assert isinstance(id[1], bool) or isinstance(id[1], str)
        
        if len(id) == 4 and id[:3] == (True, False, 1):
            # if it is the C_2 subgroup of the inversion, it is invariant to adjunction
            id = (True, False, 1, self.identity)

        if id[0] == 'cone' and len(id) == 4 and isinstance(id[1], float) and isinstance(id[2], int):
            # A cone subgroup is generally specified by ('cone', N, adj)
            # It is assumed that the flip is always along the X axis in the plane defined by adj
            # If one wants to define the flip along another axis in the plane, one can tune the adj parameter.
            # However, this can be inconvenient ofter.
            # In this case, one can also use the tuple ('cone, theta, N, adj)
            # where theta specifies 2 times the flip axis with respect to the X axis.
            # This is converted here into a tuple ('cone', N, adj) by computing the right adj value
            
            
            # The factor of 2 in id[1] comes from the fact that a flip around the axis theta, as an element of O(2), is
            # the combination of a reflection along the X axis (theta=0) and a rotation by 2*theta.
            # The values id[1] = 2*theta represents the SO(2) component of the flip.
            
            flip_axis = id[1] / 2.
            flip = np.asarray([0., 0., np.sin(flip_axis / 2), np.cos(flip_axis / 2)])
            adj = id[-1]
            adj = self.element((0, flip), 'Q') @ adj
            id = ('cone', id[2], adj)

        return id

    def _subgroup(self, id: Tuple) -> Tuple[
        Group,
        Callable[[GroupElement], GroupElement],
        Callable[[GroupElement], GroupElement]
    ]:
        r"""
        Restrict the current group :math:`O(3)` to the subgroup identified by the input ``id``

        
        Args:
            id (tuple): the identification of the subgroup

        Returns:
            a tuple containing
            
                - the subgroup
                
                - a function which maps an element of the subgroup to its inclusion in the original group and
                
                - a function which maps an element of the original group to the corresponding element in the subgroup (returns None if the element is not contained in the subgroup)
                
        """
        
        # ('so3')        -> SO(3)
        # (False, x)     -> SO(3)'s subgroup
        # (True, x)      -> SO(3)'s subgroup X Inversion
        #   includes also Cylinder as O(2) X Inversion
        # (True, None, 1) = (true, None)   -> Inversion C_2 subgroup
        # ('fulltetra')  -> full tethraedral is a semidirect prod however; it is however isomorphic to Octa group!!!
        # ('cone', x)    -> conical subgroup, O(2) or D_N depending on x

        sg_id, adj = id[:-1], id[-1]
        assert isinstance(adj, GroupElement)
        assert adj.group == self

        if sg_id == (False, 'so3'):
            # SO3 group
            sg = escnn.group.so3_group(self._maximum_frequency)
            parent_mapping = so3_to_o3(adj, sg)
            child_mapping = o3_to_so3(adj, sg)
            
        elif sg_id == (True, 'so3'):
            # the O3 group itself
            sg = self
            parent_mapping = build_adjoint_map(self, ~adj) #lambda x, adj=adj, _adj=~adj: _adj @ x @ adj
            child_mapping = build_adjoint_map(self, adj) #lambda x, adj=adj, _adj=~adj: adj @ x @ _adj

        elif sg_id[0] == 'fulltetra' and len(sg_id) == 1:
            # T_d group
            # the group is isomerphic to the Octahedral group
            # sg = Octahedral()
            raise NotImplementedError()

        elif sg_id[0] == True:
            # the rest of the id identifies a subgroup of the SO3 subgroup
            so3_sg_id = sg_id[1:]
            
            if len(so3_sg_id) == 1 and isinstance(so3_sg_id[0], str):
                if so3_sg_id[0] == 'ico':
                    # I_h
                    sg = escnn.group.full_ico_group()
                    parent_mapping = fullplato_to_o3(adj, sg)
                    child_mapping = o3_to_fullplato(adj, sg)
                elif so3_sg_id[0] == 'octa':
                    # O_h
                    sg = escnn.group.full_octa_group()
                    parent_mapping = fullplato_to_o3(adj, sg)
                    child_mapping = o3_to_fullplato(adj, sg)
                elif so3_sg_id[0] == 'tetra':
                    # T_h
                    # Pyritohedral symmetry
                    # TODO implement group
                    # sg = escnn.group.full_pyrito_group()
                    # parent_mapping = fullplato_to_o3(adj, sg)
                    # child_mapping = o3_to_fullplato(adj, sg)
                    raise NotImplementedError()
                else:
                    raise ValueError(f'Subgroup "{so3_sg_id}" not recognized!')
                
            elif len(so3_sg_id) == 2:
                # planar subgroups
                assert isinstance(so3_sg_id[1], int) or so3_sg_id[1] is None
        
                if so3_sg_id == (False, -1):
                    # SO(2) x C_2
                    sg = escnn.group.cylinder_group(self._maximum_frequency)
                    parent_mapping = cyl_to_o3(adj, sg)
                    child_mapping = o3_to_cyl(adj, sg)
                elif so3_sg_id == (True, -1):
                    # Cylinder,  O(2) x C_2
                    sg = escnn.group.full_cylinder_group(self._maximum_frequency)
                    parent_mapping = fullcyl_to_o3(adj, sg)
                    child_mapping = o3_to_fullcyl(adj, sg)
                elif so3_sg_id[0] is False and so3_sg_id[1] > 0:
                    # C_N x C_2
                    
                    N = so3_sg_id[1]
                    
                    if N == 1:
                        # Inversion subgroup C_2 of O(3)
                        sg = escnn.group.cyclic_group(2)
                        parent_mapping = inv_to_o3(self, sg)
                        child_mapping = o3_to_inv(self, sg)
                    else:
                        # C_N x C_2
                        sg = escnn.group.cylinder_discrete_group(N)
                        parent_mapping = cyl_to_o3(adj, sg)
                        child_mapping = o3_to_cyl(adj, sg)
                elif so3_sg_id[0] is True and so3_sg_id[1] > 1:
                    # Cylinder, D_N x C_2
                    N = so3_sg_id[1]
                    sg = escnn.group.full_cylinder_discrete_group(N)
                    parent_mapping = fullcyl_to_o3(adj, sg)
                    child_mapping = o3_to_fullcyl(adj, sg)
                elif so3_sg_id[0] is True and so3_sg_id[1] == 1:
                    # flip subgroup of the O(2) subgroup, C_2 x C_2
                    # This case should have already been caught in the CyclicGroup case (thanks to the
                    # process_subgroup_id method)
                    # this is equivalent to C_2 x C_2 above
                    # TODO catch this case in process_subgroup_id
                    raise ValueError(f'Subgroup "{sg_id}": this case should have already been caught!')
                else:
                    raise ValueError(f'Subgroup "{sg_id}" not recognized!')
            else:
                raise ValueError(f'Subgroup "{sg_id}" not recognized!')

        elif sg_id[0] == False:
            
            # the rest of the id identifies a subgroup of the SO3 subgroup
            so3_sg_id = sg_id[1:]
    
            if len(so3_sg_id) == 1 and isinstance(so3_sg_id[0], str):
                if so3_sg_id[0] == 'ico':
                    sg = escnn.group.ico_group()
                    parent_mapping = so3_to_o3(adj, sg)
                    child_mapping = o3_to_so3(adj, sg)
                elif so3_sg_id[0] == 'octa':
                    sg = escnn.group.octa_group()
                    parent_mapping = so3_to_o3(adj, sg)
                    child_mapping = o3_to_so3(adj, sg)
                elif so3_sg_id[0] == 'tetra':
                    # sg = escnn.group.tetra_group()
                    # parent_mapping = so3_to_o3(adj, sg)
                    # child_mapping = o3_to_so3(adj, sg)
                    raise NotImplementedError()
                else:
                    raise ValueError(f'Subgroup "{sg_id}" not recognized!')

            elif len(so3_sg_id) == 2:
                # planar subgroups
        
                if so3_sg_id == (False, -1):
                    # SO(2)
                    sg = escnn.group.so2_group(self._maximum_frequency)
                    parent_mapping = so2_to_o3(adj, sg)
                    child_mapping = o3_to_so2(adj, sg)
                elif so3_sg_id == (True, -1):
                    # Dihedral,  O(2)
                    sg = escnn.group.o2_group(self._maximum_frequency)
                    parent_mapping = dih_to_o3(adj, sg)
                    child_mapping = o3_to_dih(adj, sg)
                elif so3_sg_id[0] == False and so3_sg_id[1] > 0:
                    # Cyclic subgroup
                    # or flip subgroup of the O(2) subgroup
                    n = so3_sg_id[1]
                    assert isinstance(n, int)
                    sg = escnn.group.cyclic_group(n)
                    parent_mapping = so2_to_o3(adj, sg)
                    child_mapping = o3_to_so2(adj, sg)
        
                elif so3_sg_id[0] == True and so3_sg_id[1] > 1:
                    # Dihedral, dihedral subgroup
                    n = so3_sg_id[1]
                    assert isinstance(n, int)
                    sg = escnn.group.dihedral_group(n)
            
                    parent_mapping = dih_to_o3(adj, sg)
                    child_mapping = o3_to_dih(adj, sg)
        
                elif so3_sg_id[0] is True and so3_sg_id[1] == 1:
                    # flip subgroup of the O(2) subgroup
                    # This case should have already been caught in the CyclicGroup case (thanks to the
                    # process_subgroup_id method)
                    raise ValueError(f'Subgroup "{sg_id}": this case should have already been caught!')
                else:
                    raise ValueError(f'Subgroup "{sg_id}" not recognized!')
            else:
                raise ValueError(f'Subgroup "{sg_id}" not recognized!')

        elif sg_id[0] == 'cone':
            assert len(sg_id) == 2, sg_id
            
            N = sg_id[1]

            assert isinstance(N, int), N
            
            if N == -1:
                # O(2) cone symmetry
                sg = escnn.group.o2_group(self._maximum_frequency)
                parent_mapping = con_to_o3(adj, sg)
                child_mapping = o3_to_con(adj, sg)
            elif N > 1:
                # D_N pyramid symmetry
                sg = escnn.group.dihedral_group(N)
                parent_mapping = con_to_o3(adj, sg)
                child_mapping = o3_to_con(adj, sg)
            elif N == 1:
                # C_2 mirroring wrt a plane
                sg = escnn.group.cyclic_group(2)
                parent_mapping = mir_to_o3(adj, sg)
                child_mapping = o3_to_mir(adj, sg)

            else:
                raise ValueError(f'Subgroup "{sg_id}" not recognized!')

        else:
            raise ValueError(f'Subgroup "{sg_id}" not recognized!')

        return sg, parent_mapping, child_mapping

    def _combine_subgroups(self, sg_id1, sg_id2):
        
        sg_id1 = self._process_subgroup_id(sg_id1)
        sg1, inclusion, restriction = self.subgroup(sg_id1)
        sg_id2 = sg1._process_subgroup_id(sg_id2)
    
        sg_id1, adjoint1 = sg_id1[:-1], sg_id1[-1]
    
        adjoint2 = None
        sg_id = None
    
        if sg_id1 == (True, 'so3'):
            # subgroup of O3
            sg_id = sg_id2[:-1]
            adjoint2 = sg_id2[-1]
    
        elif sg_id1 == (False, 'so3'):
            # subgroup of SO3
            sg_id = (False,) + sg_id2[:-1]
            adjoint2 = sg_id2[-1]

        elif sg_id1[0] == 'fulltetra':
            # subgroup  of T_d group
            raise NotImplementedError()
    
        elif sg_id1[0] == True:
            # the rest of the id identifies a subgroup of the SO3 subgroup
            so3_sg_id = sg_id1[1:]
            
            # TODO :  these subgroups have the cone subgroups as subgroups
            # but since we model them as direct products, we do not implement this further subgroups operation
            # we should have a new class for cylinder symmetry whihc inheriths from DirectproductGroup and
            # which additionally implements this subgroup

            if so3_sg_id[0] in ['ico', 'octa', 'tetra']:
                # I_h
                # O_h
                # T_h:  Pyritohedral symmetry
                raise NotImplementedError()
        
            elif len(so3_sg_id) == 2:
                # planar subgroups
                assert isinstance(so3_sg_id[1], int) or so3_sg_id[1] is None
            
                if so3_sg_id == (False, -1):
                    # C_2 x SO(2)
                    sg_id = sg_id2[0], False, sg_id2[1]
            
                elif so3_sg_id == (True, -1):
                    # Cylinder,  C_2 x O(2)
                    sg_id = sg_id2[0], sg_id2[1][0], sg_id2[1][1]
            
                elif so3_sg_id[0] is False and so3_sg_id[1] > 0:
                    N = so3_sg_id[1]
                
                    if N == 1:
                        # Inversion subgroup C_2 of O(3)
                        sg_id = sg_id2, False, 1
                    else:
                        # C_2 x C_N
                        sg_id = sg_id2[0], False, sg_id2[1]
            
                elif so3_sg_id[0] is True and so3_sg_id[1] > 1:
                    # Cylinder, C_2 x D_N
                    if sg_id2[1][0] is not None:
                        sg_id = sg_id2[0], sg_id2[1][0] * 2 * np.pi / so3_sg_id[1], sg_id2[1][1]
                    else:
                        sg_id = sg_id2[0], False, sg_id2[1][1]
            
                elif so3_sg_id[0] is True and so3_sg_id[1] == 1:
                    # flip subgroup of the O(2) subgroup, C_2 x C_2
                    # This case should have already been caught in the CyclicGroup case (thanks to the
                    # process_subgroup_id method)
                    # this is equivalent to C_2 x C_2 above
                    raise ValueError(f'Subgroup "{sg_id1}": this case should have already been caught!')
                else:
                    raise ValueError(f'Subgroup "{sg_id1}" not recognized!')
            else:
                raise ValueError(f'Subgroup "{sg_id1}" not recognized!')
    
        elif sg_id1[0] == False:
        
            # the rest of the id identifies a subgroup of the SO3 subgroup
            so3_sg_id = sg1._combine_subgroups(sg_id1[1:], sg_id2)
            sg_id, adjoint2 = (False,) + so3_sg_id[:-1], so3_sg_id[-1]
    
        elif sg_id1[0] == 'cone':
            assert len(sg_id1) == 2
        
            N = sg_id1[1]
        
            assert isinstance(N, int)
        
            if N == -1 or N > 1:
                # O(2) cone symmetry
                # D_N pyramid symmetry

                if sg_id2[0] is not None:
                    sg_id = 'cone', sg_id2[1]
                    adjoint2 = sg1.element((0, sg_id2[0]))
                else:
                    sg_id = False, None, sg_id2[1]
            
            elif N == 1:
                # C_2 mirroring wrt a plane
                sg_id = sg_id1 if sg_id2 else self.subgroup_trivial_id
            else:
                raise ValueError(f'Subgroup "{sg_id}" not recognized!')
    
        else:
            raise ValueError(f'Subgroup "{sg_id}" not recognized!')
    
        if adjoint2 is not None:
            adjoint = adjoint1 @ inclusion(adjoint2)
        else:
            adjoint = adjoint1
        
        return sg_id + (adjoint,)

    def _restrict_irrep(self, irrep: Tuple, id: Tuple) -> Tuple[np.matrix, List[Tuple]]:
        r"""
        Restrict the input irrep of current group to the subgroup identified by "id".
        
        Args:
            irrep (tuple): the identifier of the irrep to restrict
            id (tuple): the identification of the subgroup

        Returns:
            a pair containing the change of basis and the list of irreps of the subgroup which appear in the restricted irrep
            
        """
    
        sg_id, adj = id[:-1], id[-1]
    
        irr = self.irrep(*irrep)
        l = irr.attributes['frequency']
        j = irr.attributes['inv_frequency']
    
        sg, _, _ = self.subgroup(id)
    
        irreps = []
        change_of_basis = None
    
        try:
            if sg_id == (True, 'so3'):
                # O3 group itself
                irreps = [irr.id]
                change_of_basis = np.eye(2*l+1)
        
            elif sg_id == (False, 'so3'):
                # SO3 group
                irreps = [(l,)]
                change_of_basis = np.eye(2*l+1)
        
            elif sg_id[0] == 'fulltetra' and len(sg_id) == 1:
                # T_d group
                # the group is isomerphic to the Octahedral group
                # sg = Octahedral()
                raise NotImplementedError()
        
            elif sg_id[0] == True:
                # the rest of the id identifies a subgroup of the SO3 subgroup
                so3_sg_id = sg_id[1:]
            
                if len(so3_sg_id) == 1 and isinstance(so3_sg_id[0], str):
                    if so3_sg_id[0] == 'ico':
                        # I_h
                        raise NotImplementedError()
                    elif so3_sg_id[0] == 'octa':
                        # O_h
                        raise NotImplementedError()
                    elif so3_sg_id[0] == 'tetra':
                        # T_h
                        # Pyritohedral symmetry
                        raise NotImplementedError()
                    else:
                        raise ValueError(f'Subgroup "{so3_sg_id}" not recognized!')
            
                elif len(so3_sg_id) == 2:
                    # planar subgroups
                    assert isinstance(so3_sg_id[1], int) or so3_sg_id[1] is None
                
                    if so3_sg_id == (False, -1):
                        # SO(2) x C_2
                        raise NotImplementedError()
                    elif so3_sg_id == (True, -1):
                        # Cylinder,  O(2) x C_2
                        raise NotImplementedError()
                    elif so3_sg_id[0] is False and so3_sg_id[1] > 0:
                        # C_N x C_2
                    
                        N = so3_sg_id[1]
                    
                        if N == 1:
                            # Inversion subgroup C_2 of O(3)
                            irreps = [(j,)]*(2*l+1)
                            change_of_basis = np.eye(2 * l + 1)
                        else:
                            # C_N x C_2
                            raise NotImplementedError()
                    elif so3_sg_id[0] is True and so3_sg_id[1] > 1:
                        # Cylinder, D_N x C_2
                        raise NotImplementedError()
                    elif so3_sg_id[0] is True and so3_sg_id[1] == 1:
                        # flip subgroup of the O(2) subgroup, C_2 x C_2
                        # This case should have already been caught in the CyclicGroup case (thanks to the
                        # process_subgroup_id method)
                        # this is equivalent to C_2 x C_2 above
                        # TODO catch this case in process_subgroup_id
                        raise ValueError(f'Subgroup "{sg_id}": this case should have already been caught!')
                    else:
                        raise ValueError(f'Subgroup "{sg_id}" not recognized!')
                else:
                    raise ValueError(f'Subgroup "{sg_id}" not recognized!')
        
            elif sg_id[0] == False:
            
                # the rest of the id identifies a subgroup of the SO3 subgroup
                so3_sg_id = sg_id[1:]
            
                # First restrict to SO(3) and then use the restriction of its irreps to the subgroup
                so3_id = self._process_subgroup_id(('so3'))
                so3, _, _ = self.subgroup(so3_id)
            
                change_of_basis_so3, irreps_so3 = self._restrict_irrep(irrep, so3_id)
            
                so3_sg_id = so3._process_subgroup_id(so3_sg_id)
            
                p = 0
                cob_so3_sg = np.zeros((2 * l + 1, 2 * l + 1))
                for irr in irreps_so3:
                    cob_sg, irr_sg = so3._restrict_irrep(irr, so3_sg_id)
                    irreps += irr_sg
                
                    irr = so3.irrep(*irr)
                    cob_so3_sg[p:p + irr.size, p:p + irr.size] = cob_sg
                    p += irr.size
            
                change_of_basis = change_of_basis_so3 @ cob_so3_sg
        
            elif sg_id[0] == 'cone':
                assert len(sg_id) == 2
            
                N = sg_id[1]
            
                assert isinstance(N, int)
            
                if N == -1:
                    # O(2) cone symmetry
                
                    irreps = [((l + j) % 2, 0)] + [(1, f) for f in range(1, l + 1)]
                    change_of_basis = np.zeros((2 * l + 1, 2 * l + 1))
                    change_of_basis[l, 0] = 1.
                    for f in range(l):
                        change_of_basis[l + f + 1, 2 * f + 1] = 1.
                        change_of_basis[l - f - 1, 2 * f + 2] = 1.
                    
                        if (l + j) % 2 == 1:
                            # TODO - prove this!
                            block = change_of_basis[[l - f - 1, l + f + 1], 2 * f + 1:2 * f + 3]
                            m = np.asarray([
                                [0., -1],
                                [1., 0.]
                            ])
                            block = m @ block
                            change_of_basis[[l - f - 1, l + f + 1], 2 * f + 1:2 * f + 3] = block
            
                elif N > 1:
                    # D_N pyramid symmetry
                    raise NotImplementedError()
                elif N == 1:
                    # C_2 mirroring wrt a plane
                    raise NotImplementedError()
                else:
                    raise ValueError(f'Subgroup "{sg_id}" not recognized!')
        
            else:
                raise ValueError(f'Subgroup "{sg_id}" not recognized!')
    
        except NotImplementedError:
            change_of_basis, irreps = restrict_irrep(irr, sg_id)

        change_of_basis = self.irrep(*irrep)(adj).T @ change_of_basis
    
        return change_of_basis, irreps

    def _build_representations(self):
        r"""
        Build the irreps for this group
        """
    
        # Build all the Irreducible Representations

        for f in range(2):
            for k in range(self._maximum_frequency + 1):
                self.irrep(f, k)
    
        # add all the irreps to the set of representations already built for this group
        self.representations.update(**{irr.name: irr for irr in self.irreps()})

    @property
    def trivial_representation(self) -> Representation:
        return self.representations['irrep_0,0']

    def standard_representation(self) -> Representation:
        r"""
        Standard representation of :math:`\O3` as 3x3 rotation matrices

        """
        name = f'standard'
    
        if name not in self._representations:
            change_of_basis = np.array([
                [0, 0, 1],
                [1, 0, 0],
                [0, 1, 0]
            ])
        
            self._representations[name] = change_basis(
                self.irrep(1, 1),
                change_of_basis=change_of_basis,
                name=name,
                supported_nonlinearities=self.irrep(1, 1).supported_nonlinearities,
            )
    
        return self._representations[name]

    def bl_regular_representation(self, L: int) -> Representation:
        r"""
        Band-Limited regular representation up to frequency ``L`` (included).

        Args:
            L(int): max frequency

        """
        irreps = []

        for l in range(L + 1):
            for j in range(2):
                irreps += [self.irrep(j, l)] * (2 * l + 1)
    
        return directsum(irreps, name=f'regular_{L}')

    def bl_quotient_representation(self,
                                   L: int,
                                   subgroup_id,
                                   ) -> escnn.group.Representation:
        r"""
        Band-Limited quotient representation up to frequency ``L`` (included).

        The quotient representation corresponds to the action of the current group :math:`G` on functions over the
        homogeneous space :math:`G/H`, where :math:`H` is the subgroup of :math:`G` identified by ``subgroup_id``.

        Args:
            L(int): max frequency
            subgroup_id: id identifying the subgroup H.

        """

        name = f"quotient[{subgroup_id}]_{L}"

        if name not in self.representations:
            subgroup, _, _ = self.subgroup(subgroup_id)

            homspace = self.homspace(subgroup_id)

            irreps = []

            for l in range(L + 1):
                for j in range(2):
                    irr = self.irrep(j, l)
                    multiplicity = homspace.dimension_basis(irr.id, homspace.H.trivial_representation.id)[1]
                    irreps += [irr] * multiplicity

            self.representations[name] = directsum(irreps, name=name)

        return self.representations[name]

    def bl_sphere_representation(self,
                                   L: int,
                                   ) -> escnn.group.Representation:
        r"""
        Representation of :math:`\O3` acting on functions on the sphere band-limited representation up to frequency
        ``L`` (included).

        Args:
            L(int): max frequency
        """
        return self.bl_quotient_representation(L, ('cone', -1))

    def irrep(self, j: int, l: int) -> IrreducibleRepresentation:
        r"""
        Build the irrep with reflection and rotation frequencies :math:`j` (reflection) and :math:`l` (rotation) of the
        current group.
        
        Args:
            j (int): the frequency of the reflection in the irrep
            l (int): the frequency of the rotations in the irrep

        Returns:
            the corresponding irrep

        """
    
        assert j in [0, 1]
        assert l >= 0
    
        name = f"irrep_{j},{l}"
        id = (j, l)

        if id not in self._irreps:

            if l == 0 and j == 0:
                # Trivial representation
                irrep = build_trivial_irrep()
                character = build_trivial_character()
                supported_nonlinearities = ['pointwise', 'norm', 'gated', 'gate']
                self._irreps[id] = IrreducibleRepresentation(self, id, name, irrep, 1, 'R',
                                                              supported_nonlinearities=supported_nonlinearities,
                                                              character=character,
                                                              frequency=0,
                                                              inv_frequency=0)
            elif l == 0 and j == 1:
                # Inversion representation
                # irrep = lambda element, identity=np.eye(1): identity * (-1 if element.to(element.group.PARAM)[0] else 1)
                # character = lambda element: (-1 if element.to(element.group.PARAM)[0] else 1)
                irrep = _build_irrep(j, l)
                character = _build_character(j, l)
                supported_nonlinearities = ['norm', 'gated']
                self._irreps[id] = IrreducibleRepresentation(self, id, name, irrep, 1, 'R',
                                                              supported_nonlinearities=supported_nonlinearities,
                                                              character=character,
                                                              frequency=0,
                                                              inv_frequency=1)
            else:

                # other Irreducible Representations
                
                # inv_action = lambda i, j=j:  (-1 if i else 1)**j
                # irrep = lambda element, l=l, inv_action=inv_action, **kwargs: _wigner_d_matrix(element.to(element.param)[1], l=l, **kwargs) * inv_action(element.to(element.param)[0])
                # character = lambda element, l=l, inv_action=inv_action, **kwargs: _character(element.to(element.param)[1], l=l, **kwargs) * inv_action(element.to(element.param)[0])
                irrep = _build_irrep(j, l)
                character = _build_character(j, l)
                supported_nonlinearities = ['norm', 'gated']
                self._irreps[id] = IrreducibleRepresentation(self, id, name, irrep, 2 * l + 1, 'R',
                                                              supported_nonlinearities=supported_nonlinearities,
                                                              character=character,
                                                              frequency=l,
                                                              inv_frequency=j)

        return self._irreps[id]

    def _tensor_product_irreps(self, J: Tuple[int, int], l: Tuple[int, int]) -> List[Tuple[Tuple, int]]:
        J = self.get_irrep_id(J)
        l = self.get_irrep_id(l)
        return [
            (((J[0] + l[0]) % 2, j), 1)
            for j in range(np.abs(J[1] - l[1]), J[1] + l[1] + 1)
        ]

    def _clebsh_gordan_coeff(self, m, n, j) -> np.ndarray:
        group_keys = self._keys
        m = self.get_irrep_id(m)
        n = self.get_irrep_id(n)
        j = self.get_irrep_id(j)

        if (m[0] + n[0])%2 == j[0]:
            return _clebsh_gordan_tensor_so3(m[1], n[1], j[1])
        else:
            return np.zeros(self.irrep(*m).size, self.irrep(*n).size, 0, self.irrep(*j).size)

    _cached_group_instance = None

    @classmethod
    def _generator(cls, maximum_frequency: int = 3) -> 'O3':
        if cls._cached_group_instance is None:
            cls._cached_group_instance = O3(maximum_frequency)
        elif cls._cached_group_instance._maximum_frequency < maximum_frequency:
            cls._cached_group_instance._maximum_frequency = maximum_frequency
            cls._cached_group_instance._build_representations()
    
        return cls._cached_group_instance


def _random_samples(N: int, seed = None, parametrization: str = PARAMETRIZATION):

    if seed is None:
        rng = np.random
    elif isinstance(seed, int):
        rng = np.random.RandomState(seed)
    else:
        assert isinstance(seed, np.random.RandomState)
        rng = seed

    so3_grid = _grid('rand', N=N, seed=rng, parametrization=parametrization)
    return [
        (rng.randint(0, 2), g) for g in so3_grid
    ]


def _build_irrep(j: int, l: int):

    def irrep(e: GroupElement, j: int = j, l: int = l):
        inversion, rotation = e.value

        inv_action = (-1 if inversion else 1) ** j

        if l == 0:
            return np.asarray([[inv_action]], dtype=np.float)
        else:
            return _wigner_d_matrix(rotation, l=l, param=e.param) * inv_action

    return irrep


def _build_character(j: int, l: int):
    
    def character(e: GroupElement, j: int = j, l: int = l):
        inversion, rotation = e.value
    
        inv_action = (-1 if inversion else 1) ** j
        if l == 0:
            return inv_action
        else:
            return _character(rotation, l, param=e.param) * inv_action
    
    return character


#############################################
# SUBGROUPS MAPS
#############################################

# Full Icosahedral I_h, Full Octahedral O_h or Pyritohedral T_h (not Full Tetra!!) ###############################

def o3_to_fullplato(adj: GroupElement, plato: escnn.group.DirectProductGroup):
    assert isinstance(adj.group, O3)
    
    def _map(e: GroupElement, plato=plato, adj=adj):
        o3 = adj.group
        assert e.group == o3

        try:
            return plato.element(
                (adj @ e @ (~adj)).to('Q'), # = (flip, rot)
                '[int | Q]'
            )
        except ValueError:
            return None
    
    return _map


def fullplato_to_o3(adj: GroupElement, plato: escnn.group.DirectProductGroup):
    assert isinstance(adj.group, O3)
    
    def _map(e: GroupElement, plato=plato, adj=adj):
        assert e.group == plato
        o3 = adj.group
        return (~adj) @ o3.element(e.to('[int | Q]'), 'Q') @ adj
    
    return _map


# SO3 (and Icosahedral, Octahedral, Tetrtahedral) ###############################

def o3_to_so3(adj: GroupElement, so3: Union[escnn.group.SO3, escnn.group.Icosahedral]):
    # can also support Tetra and Octa
    
    assert isinstance(adj.group, O3)
    
    def _map(e: GroupElement, so3=so3, adj=adj):
        o3 = adj.group
        assert e.group == o3
        
        flip, rotation = (adj @ e @ (~adj)).to('Q')
        
        if flip == 0:
            try:
                return so3.element(rotation, 'Q')
            except ValueError:
                return None
        else:
            return None
    
    return _map


def so3_to_o3(adj: GroupElement, so3: Union[escnn.group.SO3, escnn.group.Icosahedral]):
    # can also support Tetra and Octa
    
    assert isinstance(adj.group, O3)
    
    def _map(e: GroupElement, so3=so3, adj=adj):
        assert e.group == so3
        o3 = adj.group
        return (~adj) @ o3.element(
            (0, e.to('Q')),
            'Q'
        ) @ adj
    
    return _map


# SO(2) (and C_N) #####################################

def o3_to_so2(adj: GroupElement, so2: Union[escnn.group.SO2, escnn.group.CyclicGroup]):
    assert isinstance(adj.group, O3)
    
    def _map(e: GroupElement, so2=so2, adj=adj):
        o3 = adj.group
        assert e.group == o3
        
        e = adj @ e @ (~adj)
        
        flip, e = e.to('Q')
        
        if flip == 1:
            return None
        
        if np.allclose(e[:2], 0.):
            # if the rotation is along the Z axis, i.e. on the XY plane
            s, c = e[2:]
            theta = 2 * np.arctan2(s, c)
            try:
                return so2.element(theta, 'radians')
            except ValueError:
                return None
        else:
            return None
    
    return _map


def so2_to_o3(adj: GroupElement, so2: Union[escnn.group.SO2, escnn.group.CyclicGroup]):
    assert isinstance(adj.group, O3)
    
    def _map(e: GroupElement, so2=so2, adj=adj):
        assert e.group == so2
        o3 = adj.group
        
        theta_2 = e.to('radians') / 2.
        
        q = np.asarray([0., 0., np.sin(theta_2), np.cos(theta_2)])
        
        return (~adj) @ o3.element((0, q), 'Q') @ adj
    
    return _map


# Dihedral (O(2) and D_N) ######################################

def o3_to_dih(adj: GroupElement, dih: Union[escnn.group.O2, escnn.group.DihedralGroup]):
    assert isinstance(adj.group, O3)
    
    def _map(e: GroupElement, dih=dih, adj=adj):
        o3 = adj.group
        assert e.group == o3
        
        e = adj @ e @ (~adj)
        
        flip, e = e.to('Q')
        
        if flip == 1:
            return None

        if np.allclose(e[:2], 0.):
            # if it is a rotation along the Z axis, i.e. on the XY plane
            s, c = e[2:]
            theta = 2 * np.arctan2(s, c)
            flip = 0
            try:
                return dih.element((flip, theta), 'radians')
            except ValueError:
                return None
        elif np.allclose(e[2:], 0.):
            # if it is a rotation by 180 degrees around an axis perpendicular to Z, i.e. an axis in the XY plane
            c, s = e[:2]
            theta = 2 * np.arctan2(s, c)
            flip = 1
            try:
                return dih.element((flip, theta), 'radians')
            except ValueError:
                return None
        else:
            return None
    
    return _map


def dih_to_o3(adj: GroupElement, dih: Union[escnn.group.O2, escnn.group.DihedralGroup]):
    assert isinstance(adj.group, O3)
    
    def _map(e: GroupElement, dih=dih, adj=adj):
        assert e.group == dih
        o3 = adj.group
        
        f, theta = e.to('radians')
        
        theta_2 = theta / 2.
        s, c = np.sin(theta_2), np.cos(theta_2)
        
        if f == 0:
            q = np.asarray([0., 0., s, c])
        elif f == 1:
            q = np.asarray([c, s, 0., 0.])
        else:
            raise ValueError()
        
        return (~adj) @ o3.element((0, q), 'Q') @ adj
    
    return _map


# Conical (O(2) and D_N) ######################################


def o3_to_con(adj: GroupElement, con: Union[escnn.group.O2, escnn.group.DihedralGroup]):
    assert isinstance(adj.group, O3)

    o3 = adj.group
    reflection = o3.element(
        (1, np.asarray([0., 1., 0., 0.])),
        'Q'
    )

    def _map(e: GroupElement, con=con, adj=adj, reflection=reflection):
        o3 = adj.group
        assert e.group == o3
        
        e = adj @ e @ (~adj)

        _, p = (e @ reflection).to('Q')

        flip, e = e.to('Q')
        
        if flip == 0 and np.allclose(e[:2], 0.):
            # if it is a rotation along the Z axis, i.e. on the XY plane
            s, c = e[2:]
            theta = 2 * np.arctan2(s, c)
            try:
                return con.element((0, theta), 'radians')
            except ValueError:
                return None
        elif flip == 1 and np.allclose(p[:2], 0.):
            # if it is a reflection along a plane perpendicular to the XY plane
            s, c = p[2:]
            theta = 2 * np.arctan2(s, c)
            try:
                return con.element((1, theta), 'radians')
            except ValueError:
                return None

        else:
            return None
    
    return _map


def con_to_o3(adj: GroupElement, con: Union[escnn.group.O2, escnn.group.DihedralGroup]):
    assert isinstance(adj.group, O3)
    
    o3 = adj.group
    reflection = o3.element(
        (1, np.asarray([0., 1., 0., 0.])),
        'Q'
    )

    def _map(e: GroupElement, con=con, adj=adj, reflection=reflection):
        assert e.group == con
        o3 = adj.group
        
        f, theta = e.to('radians')
        
        theta_2 = theta / 2.
        s, c = np.sin(theta_2), np.cos(theta_2)

        e = o3.element(
            (0, np.asarray([0., 0., s, c])),
            'Q'
        )
        
        if f == 1:
            e = e @ (~reflection)
            
        return (~adj) @ e @ adj
    
    return _map


# Mirroring wrt a plane ######################################


def o3_to_mir(adj: GroupElement, mir: escnn.group.CyclicGroup):
    
    assert isinstance(adj.group, O3)
    assert isinstance(mir, escnn.group.CyclicGroup) and mir.order() == 2

    o3 = adj.group
    reflection = o3.element(
        (1, np.asarray([0., 1., 0., 0.])),
        'Q'
    )
    
    def _map(e: GroupElement, mir=mir, adj=adj, reflection=reflection):
        o3 = adj.group
        assert e.group == o3
        
        e = adj @ e @ (~adj)
        
        _, p = (e @ reflection).to('Q')
        
        flip, e = e.to('Q')
        
        if flip == 0 and np.allclose(e[:3], 0.):
            # if it is the identity
            return mir.identity
        elif flip == 1 and np.allclose(p[:3], 0.):
            # if it is a mirroring wrt the XZ plane, i.e. along the Y axis
            return mir.element(1, 'int')
        else:
            return None
    
    return _map


def mir_to_o3(adj: GroupElement, mir: escnn.group.CyclicGroup):
    
    assert isinstance(adj.group, O3)
    assert isinstance(mir, escnn.group.CyclicGroup) and mir.order() == 2

    o3 = adj.group
    reflection = o3.element(
        (1, np.asarray([0., 1., 0., 0.])),
        'Q'
    )
    
    def _map(e: GroupElement, mir=mir, adj=adj, reflection=reflection):
        assert e.group == mir
        o3 = adj.group
        
        f = e.to('int')
        
        if f == 0:
            return o3.identity
        else:
            return (~adj) @ reflection @ adj
    
    return _map


# Inversions subgroup ###############################

def o3_to_inv(o3: O3, inv: escnn.group.CyclicGroup):
    
    assert isinstance(inv, escnn.group.CyclicGroup) and inv.order() == 2
    
    def _map(e: GroupElement, inv=inv, o3=o3):
        
        assert e.group == o3

        flip, rotation = e.to('Q')
        
        if np.allclose(rotation[:3], 0.) and np.allclose(np.abs(rotation[3]), 1.):
            return inv.element(flip, 'int')
        else:
            return None
    
    return _map


def inv_to_o3(o3: O3, inv: escnn.group.CyclicGroup):
    
    assert isinstance(inv, escnn.group.CyclicGroup) and inv.order() == 2
    
    def _map(e: GroupElement, inv=inv, o3=o3):
        assert e.group == inv
        
        flip = e.to('int')
        
        return o3.element(
            (flip, np.array([0., 0., 0., 1.])),
            'Q'
        )
    
    return _map


# Cylinder C_2 x SO(2) (and C_2 x C_N) #####################################

def o3_to_cyl(adj: GroupElement, cyl: escnn.group.DirectProductGroup):
    assert isinstance(adj.group, O3)
    assert isinstance(cyl.G1, escnn.group.CyclicGroup) and cyl.G1.order() == 2
    assert isinstance(cyl.G2, escnn.group.CyclicGroup) or isinstance(cyl.G2, escnn.group.SO2)

    def _map(e: GroupElement, cyl=cyl, adj=adj):
        o3 = adj.group
        assert e.group == o3

        e = adj @ e @ (~adj)

        inv, e = e.to('Q')

        if np.allclose(e[:2], 0.):
            # if the rotation is along the Z axis, i.e. on the XY plane
            s, c = e[2:]
            theta = 2 * np.arctan2(s, c)
            try:
                return cyl.element((inv, theta), '[int | radians]')
            except ValueError:
                return None
        else:
            return None

    return _map


def cyl_to_o3(adj: GroupElement, cyl: escnn.group.DirectProductGroup):
    assert isinstance(adj.group, O3)
    assert isinstance(cyl.G1, escnn.group.CyclicGroup) and cyl.G1.order() == 2
    assert isinstance(cyl.G2, escnn.group.CyclicGroup) or isinstance(cyl.G2, escnn.group.SO2)

    def _map(e: GroupElement, cyl=cyl, adj=adj):
        assert e.group == cyl
        o3 = adj.group

        inv, theta = e.to('[int | radians]')
        theta_2 = theta / 2.

        q = np.asarray([0., 0., np.sin(theta_2), np.cos(theta_2)])

        return (~adj) @ o3.element((inv, q), 'Q') @ adj

    return _map


# Full Cylinder C_2 x O(2) (and C_2 x D_N) #####################################

def o3_to_fullcyl(adj: GroupElement, fullcyl: escnn.group.DirectProductGroup):
    assert isinstance(adj.group, O3)
    assert isinstance(fullcyl.G1, escnn.group.CyclicGroup) and fullcyl.G1.order() == 2
    assert isinstance(fullcyl.G2, escnn.group.DihedralGroup) or isinstance(fullcyl.G2, escnn.group.O2)

    def _map(e: GroupElement, fullcyl=fullcyl, adj=adj):
        o3 = adj.group
        assert e.group == o3

        e = adj @ e @ (~adj)

        inv, e = e.to('Q')

        if np.allclose(e[:2], 0.):
            # if it is a rotation along the Z axis, i.e. on the XY plane
            s, c = e[2:]
            theta = 2 * np.arctan2(s, c)
            flip = 0
            try:
                return fullcyl.element((inv, (flip, theta)), '[int | radians]')
            except ValueError:
                return None
        elif np.allclose(e[2:], 0.):
            # if it is a rotation by 180 degrees around an axis perpendicular to Z, i.e. an axis in the XY plane
            c, s = e[:2]
            theta = 2 * np.arctan2(s, c)
            flip = 1
            try:
                return fullcyl.element((inv, (flip, theta)), '[int | radians]')
            except ValueError:
                return None
        else:
            return None

    return _map


def fullcyl_to_o3(adj: GroupElement, fullcyl: escnn.group.DirectProductGroup):
    assert isinstance(adj.group, O3)
    assert isinstance(fullcyl.G1, escnn.group.CyclicGroup) and fullcyl.G1.order() == 2
    assert isinstance(fullcyl.G2, escnn.group.DihedralGroup) or isinstance(fullcyl.G2, escnn.group.O2)

    def _map(e: GroupElement, fullcyl=fullcyl, adj=adj):
        assert e.group == fullcyl
        o3 = adj.group

        inv, el_o2 = e.to('[int | radians]')
        f, theta = el_o2

        theta_2 = theta / 2.
        s, c = np.sin(theta_2), np.cos(theta_2)

        if f == 0:
            q = np.asarray([0., 0., s, c])
        elif f == 1:
            q = np.asarray([c, s, 0., 0.])
        else:
            raise ValueError()

        return (~adj) @ o3.element((inv, q), 'Q') @ adj

    return _map

############################################




from __future__ import annotations

import escnn.kernels
import escnn.group
from escnn.group import Group
from .gspace import GSpace

from typing import Tuple, Callable

import numpy as np


__all__ = [
    "GSpace0D",
    "no_base_space",
]


class GSpace0D(GSpace):
    
    def __init__(self, G: escnn.group.Group):
        # TODO: is GSpace0D a good name? Maybe RepresentationSpace is better?
        super(GSpace0D, self).__init__(G, 0, G.name)

    def restrict(self, id) -> Tuple[GSpace, Callable, Callable]:
        subgroup, parent, child = self.fibergroup.subgroup(id)
        return GSpace0D(subgroup), parent, child

    @property
    def basespace_action(self) -> escnn.group.Representation:
        return None
    
    def _interpolate_transform_basespace(
            self,
            input: np.ndarray,
            element: escnn.group.GroupElement,
            order: int = 2,
    ) -> np.ndarray:
        assert element.group == self.fibergroup
        return input

    def build_kernel_basis(self,
                           in_repr: escnn.group.Representation,
                           out_repr: escnn.group.Representation,
                           **kwargs) -> escnn.kernels.KernelBasis:
        r"""

        Args:
            in_repr (Representation): the input representation
            out_repr (Representation): the output representation
            **kwargs: Group-specific keywords arguments for ``_basis_generator`` method

        Returns:
            the analytical basis

        """
        assert isinstance(in_repr, escnn.group.Representation)
        assert isinstance(out_repr, escnn.group.Representation)

        assert in_repr.group == self.fibergroup
        assert out_repr.group == self.fibergroup

        # build the key
        key = ()

        if (in_repr.name, out_repr.name) not in self._fields_intertwiners_basis_memory[key]:
    
            basis = self._basis_generator(in_repr, out_repr)
    
            # store the basis in the dictionary
            self._fields_intertwiners_basis_memory[key][(in_repr.name, out_repr.name)] = basis

        # return the dictionary with the basis built
        return self._fields_intertwiners_basis_memory[key][(in_repr.name, out_repr.name)]

    def _basis_generator(self,
                         in_repr: escnn.group.Representation,
                         out_repr: escnn.group.Representation,
                         **kwargs):
        return escnn.kernels.kernels_on_point(in_repr, out_repr)

    def __eq__(self, other):
        if isinstance(other, GSpace0D):
            return self.fibergroup == other.fibergroup
        else:
            return False

    def __hash__(self):
        return hash(self.fibergroup)


def no_base_space(G: Group) -> GSpace0D:
    r"""
    
    Build the :class:`~escnn.gspaces.GSpace` of the input group ``G`` acting on a single point space.
    
    This simple gspace can be useful to describe the features of a ``G``-equivariant MLP.
    
    Args:
        G (Group): a group

    """
    return GSpace0D(G)



from abc import ABC, abstractmethod

from escnn.nn import FieldType
from escnn.nn import GeometricTensor
from escnn.group import Representation
from escnn.kernels import KernelBasis
from escnn.gspaces import *

from ..equivariant_module import EquivariantModule

from escnn.nn.modules.basismanager import BasisManager
from escnn.nn.modules.basismanager import BlocksBasisExpansion

from typing import Callable, Union, Tuple, List

import torch
from torch.nn import Parameter
import numpy as np
import math


__all__ = ["_RdConv"]


class _RdConv(EquivariantModule, ABC):
    
    def __init__(self,
                 in_type: FieldType,
                 out_type: FieldType,
                 d: int,
                 kernel_size: int,
                 padding: int = 0,
                 stride: int = 1,
                 dilation: int = 1,
                 padding_mode: str = 'zeros',
                 groups: int = 1,
                 bias: bool = True,
                 basis_filter: Callable[[dict], bool] = None,
                 recompute: bool = False,
                 ):
        r"""

        Abstract class which implements a general G-steerable convolution, mapping between the input and output
        :class:`~escnn.nn.FieldType` s specified by the parameters ``in_type`` and ``out_type``.
        This operation is equivariant under the action of :math:`\R^d\rtimes G` where :math:`G` is the
        :attr:`escnn.nn.FieldType.fibergroup` of ``in_type`` and ``out_type``.
        
        Specifically, let :math:`\rho_\text{in}: G \to \GL{\R^{c_\text{in}}}` and
        :math:`\rho_\text{out}: G \to \GL{\R^{c_\text{out}}}` be the representations specified by the input and output
        field types.
        Then :class:`~escnn.nn._RdConv` guarantees an equivariant mapping
        
        .. math::
            \kappa \star [\mathcal{T}^\text{in}_{g,u} . f] = \mathcal{T}^\text{out}_{g,u} . [\kappa \star f] \qquad\qquad \forall g \in G, u \in \R^d
            
        where the transformation of the input and output fields are given by
 
        .. math::
            [\mathcal{T}^\text{in}_{g,u} . f](x) &= \rho_\text{in}(g)f(g^{-1} (x - u)) \\
            [\mathcal{T}^\text{out}_{g,u} . f](x) &= \rho_\text{out}(g)f(g^{-1} (x - u)) \\

        The equivariance of G-steerable convolutions is guaranteed by restricting the space of convolution kernels to an
        equivariant subspace.
        As proven in `3D Steerable CNNs <https://arxiv.org/abs/1807.02547>`_, this parametrizes the *most general
        equivariant convolutional map* between the input and output fields.

        .. warning ::

            This class implements a *discretized* convolution operator over a discrete grid.
            This means that equivariance to continuous symmetries is *not* perfect.
            In practice, by using sufficiently band-limited filters, the equivariance error introduced by the
            discretization of the filters and the features is contained, but some design choices may have a negative
            effect on the overall equivariance of the architecture.

        During training, in each forward pass the module expands the basis of G-steerable kernels with learned weights
        before performing the convolution.
        When :meth:`~torch.nn.Module.eval()` is called, the filter is built with the current trained weights and stored
        for future reuse such that no overhead of expanding the kernel remains.
        
        .. warning ::
            
            When :meth:`~torch.nn.Module.train()` is called, the attributes :attr:`~escnn.nn.R2Conv.filter` and
            :attr:`~escnn.nn.R2Conv.expanded_bias` are discarded to avoid situations of mismatch with the
            learnable expansion coefficients.
            See also :meth:`escnn.nn._RdConv.train`.
            
            This behaviour can cause problems when storing the :meth:`~torch.nn.Module.state_dict` of a model while in
            a mode and lately loading it in a model with a different mode, as the attributes of the class change.
            To avoid this issue, we recommend converting the model to eval mode before storing or loading the state
            dictionary.
 
        
        Args:
            in_type (FieldType): the type of the input field, specifying its transformation law
            out_type (FieldType): the type of the output field, specifying its transformation law
            d (int): dimensionality of the base space (2 for images, 3 for volumes)
            kernel_size (int): the size of the (square) filter
            padding(int, optional): implicit zero paddings on both sides of the input. Default: ``0``
            padding_mode(str, optional): ``zeros``, ``reflect``, ``replicate`` or ``circular``. Default: ``zeros``
            stride(int, optional): the stride of the kernel. Default: ``1``
            dilation(int, optional): the spacing between kernel elements. Default: ``1``
            groups (int, optional): number of blocked connections from input channels to output channels.
                                    It allows depthwise convolution. When used, the input and output types need to be
                                    divisible in ``groups`` groups, all equal to each other.
                                    Default: ``1``.
            bias (bool, optional): Whether to add a bias to the output (only to fields which contain a
                    trivial irrep) or not. Default ``True``
            basis_filter (callable, optional): filter for the basis elements. Should take a dictionary containing an
                                               element's attributes and return whether to keep it or not.
            recompute (bool, optional): if ``True``, recomputes a new basis for the equivariant kernels.
                    By Default (``False``), it  caches the basis built or reuse a cached one, if it is found.
        
        Attributes:
            
            ~.weights (torch.Tensor): the learnable parameters which are used to expand the kernel
            ~.filter (torch.Tensor): the convolutional kernel obtained by expanding the parameters
                                    in :attr:`~escnn.nn.R2Conv.weights`
            ~.bias (torch.Tensor): the learnable parameters which are used to expand the bias, if ``bias=True``
            ~.expanded_bias (torch.Tensor): the equivariant bias which is summed to the output, obtained by expanding
                                    the parameters in :attr:`~escnn.nn.R2Conv.bias`
        
        """

        assert in_type.gspace == out_type.gspace
        assert isinstance(in_type.gspace, GSpace)
        assert d >= in_type.gspace.dimensionality

        super(_RdConv, self).__init__()

        self.d = d
        self.space = in_type.gspace
        self.in_type = in_type
        self.out_type = out_type
        
        self.kernel_size = kernel_size
        self.stride = stride
        self.dilation = dilation
        self.padding = padding
        self.padding_mode = padding_mode
        self.groups = groups

        if isinstance(padding, tuple) and len(padding) == self.d:
            _padding = padding
        elif isinstance(padding, int):
            _padding = (padding,)*self.d
        else:
            raise ValueError('padding needs to be either an integer or a tuple containing {} integers but {} found'.format(self.d, padding))
        
        padding_modes = {'zeros', 'reflect', 'replicate', 'circular'}
        if padding_mode not in padding_modes:
            raise ValueError("padding_mode must be one of [{}], but got padding_mode='{}'".format(padding_modes, padding_mode))
        self._reversed_padding_repeated_twice = tuple(x for x in reversed(_padding) for _ in range(self.d))
        
        if groups > 1:
            # Check the input and output classes can be split in `groups` groups, all equal to each other
            # first, check that the number of fields is divisible by `groups`
            assert len(in_type) % groups == 0
            assert len(out_type) % groups == 0
            in_size = len(in_type) // groups
            out_size = len(out_type) // groups
            
            # then, check that all groups are equal to each other, i.e. have the same types in the same order
            assert all(in_type.representations[i] == in_type.representations[i % in_size] for i in range(len(in_type)))
            assert all(out_type.representations[i] == out_type.representations[i % out_size] for i in range(len(out_type)))
            
            # finally, retrieve the type associated to a single group in input.
            # this type will be used to build a smaller kernel basis and a smaller filter
            # as in PyTorch, to build a filter for grouped convolution, we build a filter which maps from one input
            # group to all output groups. Then, PyTorch's standard convolution routine interpret this filter as `groups`
            # different filters, each mapping an input group to an output group.
            in_type = in_type.index_select(list(range(in_size)))
        
        if bias:
            # bias can be applied only to trivial irreps inside the representation
            # to apply bias to a field we learn a bias for each trivial irreps it contains
            # and, then, we transform it with the change of basis matrix to be able to apply it to the whole field
            # this is equivalent to transform the field to its irreps through the inverse change of basis,
            # sum the bias only to the trivial irrep and then map it back with the change of basis
            
            # count the number of trivial irreps
            trivials = 0
            for r in self.out_type:
                for irr in r.irreps:
                    if self.out_type.fibergroup.irrep(*irr).is_trivial():
                        trivials += 1
            
            # if there is at least 1 trivial irrep
            if trivials > 0:
                
                # matrix containing the columns of the change of basis which map from the trivial irreps to the
                # field representations. This matrix allows us to map the bias defined only over the trivial irreps
                # to a bias for the whole field more efficiently
                bias_expansion = torch.zeros(self.out_type.size, trivials)
                
                p, c = 0, 0
                for r in self.out_type:
                    pi = 0
                    for irr in r.irreps:
                        irr = self.out_type.fibergroup.irrep(*irr)
                        if irr.is_trivial():
                            bias_expansion[p:p+r.size, c] = torch.tensor(r.change_of_basis[:, pi])
                            c += 1
                        pi += irr.size
                    p += r.size
                
                self.register_buffer("bias_expansion", bias_expansion)
                self.bias = Parameter(torch.zeros(trivials), requires_grad=True)
                self.register_buffer("expanded_bias", torch.zeros(out_type.size))
            else:
                self.bias = None
                self.expanded_bias = None
        else:
            self.bias = None
            self.expanded_bias = None

        # compute the coordinates of the centers of the cells in the grid where the filter is sampled
        grid = get_grid_coords(d, kernel_size, dilation)

        # note that `in_type` is used instead of `self.in_type` such that it works also when `groups > 1`
        
        # BlocksBasisExpansion: submodule which takes care of building the filter
        self._basisexpansion = BlocksBasisExpansion(in_type.representations, out_type.representations,
                                                    self._build_kernel_basis,
                                                    grid,
                                                    basis_filter=basis_filter,
                                                    recompute=recompute)
        
        if self.basisexpansion.dimension() == 0:
            raise ValueError('''
                The basis for the steerable filter is empty!
                Tune the `frequencies_cutoff`, `kernel_size`, `rings`, `sigma` or `basis_filter` parameters to allow
                for a larger basis.
            ''')
        
        self.weights = Parameter(torch.zeros(self.basisexpansion.dimension()), requires_grad=True)
        
        filter_size = (out_type.size, in_type.size) + (kernel_size,) * d
        self.register_buffer("filter", torch.zeros(*filter_size))
    
    @abstractmethod
    def _build_kernel_basis(self, in_repr: Representation, out_repr: Representation) -> KernelBasis:
        raise NotImplementedError
    
    @property
    def basisexpansion(self) -> BlocksBasisExpansion:
        r"""
        Submodule which takes care of building the filter.
        
        It uses the learnt ``weights`` to expand a basis and returns a filter in the usual form used by conventional
        convolutional modules.
        It uses the learned ``weights`` to expand the kernel in the G-steerable basis and returns it in the shape
        :math:`(c_\text{out}, c_\text{in}, s^d)`, where :math:`s` is the ``kernel_size`` and :math:`d` is the
        dimensionality of the base space.
        
        """
        return self._basisexpansion
    
    def expand_parameters(self) -> Tuple[torch.Tensor, torch.Tensor]:
        r"""
        
        Expand the filter in terms of the :attr:`~escnn.nn._RdConv.weights` and the
        expanded bias in terms of :class:`~escnn.nn._RdConv.bias`.
        
        Returns:
            the expanded filter and bias

        """
        _filter = self.basisexpansion(self.weights)
        _filter = _filter.reshape(_filter.shape[0], _filter.shape[1], *(self.kernel_size,)*self.d)
        
        if self.bias is None:
            _bias = None
        else:
            _bias = self.bias_expansion @ self.bias
            
        return _filter, _bias
    
    @abstractmethod
    def forward(self, input: GeometricTensor):
        r"""
        Convolve the input with the expanded filter and bias.
        
        Args:
            input (GeometricTensor): input feature field transforming according to ``in_type``

        Returns:
            output feature field transforming according to ``out_type``
            
        """
        pass
        
    def train(self, mode=True):
        r"""
        
        If ``mode=True``, the method sets the module in training mode and discards the :attr:`~escnn.nn._RdConv.filter`
        and :attr:`~escnn.nn._RdConv.expanded_bias` attributes.
        
        If ``mode=False``, it sets the module in evaluation mode. Moreover, the method builds the filter and the bias
        using the current values of the trainable parameters and store them in :attr:`~escnn.nn._RdConv.filter` and
        :attr:`~escnn.nn._RdConv.expanded_bias` such that they are not recomputed at each forward pass.
        
        .. warning ::
            
            This behaviour can cause problems when storing the :meth:`~torch.nn.Module.state_dict` of a model while in
            a mode and lately loading it in a model with a different mode, as the attributes of this class change.
            To avoid this issue, we recommend converting the model to eval mode before storing or loading the state
            dictionary.
        
        Args:
            mode (bool, optional): whether to set training mode (``True``) or evaluation mode (``False``).
                                   Default: ``True``.

        """

        if mode:
            # TODO thoroughly check this is not causing problems
            if hasattr(self, "filter"):
                del self.filter
            if hasattr(self, "expanded_bias"):
                del self.expanded_bias
        elif self.training:
            # avoid re-computation of the filter and the bias on multiple consecutive calls of `.eval()`
    
            _filter, _bias = self.expand_parameters()
    
            self.register_buffer("filter", _filter)
            if _bias is not None:
                self.register_buffer("expanded_bias", _bias)
            else:
                self.expanded_bias = None

        return super(_RdConv, self).train(mode)

    def evaluate_output_shape(self, input_shape: Tuple) -> Tuple:
        assert len(input_shape) == 2 + self.d
        assert input_shape[1] == self.in_type.size
    
        b, c = input_shape[:2]
        w = input_shape[2:]
        
        wo = [None]*self.d
        for i in range(self.d):
            wo[i] = math.floor((w[i] + 2 * self.padding - self.dilation * (self.kernel_size - 1) - 1) / self.stride + 1)

        return (b, self.out_type.size) + tuple(wo)

    def __repr__(self):
        extra_lines = []
        extra_repr = self.extra_repr()
        if extra_repr:
            extra_lines = extra_repr.split('\n')

        main_str = self._get_name() + '('
        if len(extra_lines) == 1:
            main_str += extra_lines[0]
        else:
            main_str += '\n  ' + '\n  '.join(extra_lines) + '\n'

        main_str += ')'
        return main_str

    def extra_repr(self):
        s = ('{in_type}, {out_type}, kernel_size={kernel_size}, stride={stride}')
        if self.padding != 0 and self.padding != (0,)*self.d:
            s += ', padding={padding}'
        if self.dilation != 1 and self.dilation != (1,)*self.d:
            s += ', dilation={dilation}'
        if self.groups != 1:
            s += ', groups={groups}'
        if self.bias is None:
            s += ', bias=False'
        return s.format(**self.__dict__)


def get_grid_coords(d: int, kernel_size: int, dilation: int = 1) -> np.ndarray:
    
    actual_size = dilation * (kernel_size -1) + 1
    
    origin = actual_size / 2 - 0.5
    
    points = np.empty((kernel_size**d, d))
    
    for i in range(kernel_size**d):
        
        for j in range(d):
            points[i, j] = (i // (kernel_size**j)) % kernel_size
            points[i, j] *= dilation
            
            # center the origin
            points[i, j] -= origin

            if j >= 1:
                # invert Y and Z coordinates
                # TODO : should this hold also for other coordinates in R^d, d > 3?
                points[i, j] *= -1
        
    return points




from escnn.nn import FieldType

from escnn.gspaces import GSpace3D
from escnn.group import Representation, Group, Icosahedral
from escnn.kernels import KernelBasis
from escnn.kernels import kernels_aliased_Ico_act_R3_icosahedron, kernels_aliased_Ico_act_R3_icosidodecahedron, kernels_aliased_Ico_act_R3_dodecahedron

from .r3convolution import R3Conv

from typing import Callable, Union, List


__all__ = ["R3IcoConv"]


class R3IcoConv(R3Conv):
    
    def __init__(
            self,
            in_type: FieldType,
            out_type: FieldType,
            kernel_size: int,
            padding: int = 0,
            stride: int = 1,
            dilation: int = 1,
            padding_mode: str = 'zeros',
            groups: int = 1,
            bias: bool = True,
            samples: str = 'ico',
            sigma: Union[List[float], float] = None,
            rings: List[float] = None,
            recompute: bool = False,
            basis_filter: Callable[[dict], bool] = None,
            initialize: bool = True,
    ):
        r"""


        Icosahedral-steerable volumetric convolution mapping between the input and output :class:`~escnn.nn.FieldType` s specified by
        the parameters ``in_type`` and ``out_type``.
        This operation is equivariant under the action of :math:`\R^3\rtimes I` where :math:`I`
        (the :class:`~escnn.group.Icosahedral` group) is the :attr:`escnn.nn.FieldType.fibergroup` of
        ``in_type`` and ``out_type``.

        This class is mostly similar to :class:`~escnn.nn.R3Conv`, with the only difference that it only supports the
        group :class:`~escnn.group.Icosahedral` since it uses a kernel basis which is specific for this group.

        The argument ``frequencies_cutoff`` of :class:`~escnn.nn.R3Conv` is not supported here since the steerable
        kernels are not generated from a band-limited set of harmonic functions.

        Instead, the argument ``samples`` specifies the polyhedron (symmetric with respect to the
        :class:`~escnn.group.Icosahedral`  group) whose vertices are used to define the kernel on :math:`\R^3`.
        The supported polyhedrons are ``"ico"`` (the 12 vertices of the icosahedron), ``"dodeca"`` (the 20 vertices
        of the dodecahedron) or ``"icosidodeca"`` (the 30 vertices of the icosidodecahedron, which correspond to the
        centers of the 30 edges of either the icosahedron or the dodecahedron).

        For each ring ``r`` in ``rings``, the polyhedron specified in embedded in the sphere of radius ``r``.
        The analytical kernel, which is only defined on the vertices of this polyhedron, is then "diffused" in the
        ambient space :math:`\R^3` by means of a small Gaussian kernel with std ``sigma``.

        .. warning ::
            Even if the input tensor has a `coords` attribute, the output of this module will not have one.

        Args:
            in_type (FieldType): the type of the input field, specifying its transformation law
            out_type (FieldType): the type of the output field, specifying its transformation law
            kernel_size (int): the size of the (square) filter
            padding(int, optional): implicit zero paddings on both sides of the input. Default: ``0``
            stride(int, optional): the stride of the kernel. Default: ``1``
            dilation(int, optional): the spacing between kernel elements. Default: ``1``
            padding_mode(str, optional): ``zeros``, ``reflect``, ``replicate`` or ``circular``. Default: ``zeros``
            groups (int, optional): number of blocked connections from input channels to output channels.
                                    It allows depthwise convolution. When used, the input and output types need to be
                                    divisible in ``groups`` groups, all equal to each other.
                                    Default: ``1``.
            bias (bool, optional): Whether to add a bias to the output (only to fields which contain a
                    trivial irrep) or not. Default ``True``
            sigma (list or float, optional): width of each ring where the bases are sampled. If only one scalar
                    is passed, it is used for all rings.
            rings (list, optional): radii of the rings where to sample the bases
            recompute (bool, optional): if ``True``, recomputes a new basis for the equivariant kernels.
                    By Default (``False``), it  caches the basis built or reuse a cached one, if it is found.
            basis_filter (callable, optional): function which takes as input a descriptor of a basis element
                    (as a dictionary) and returns a boolean value: whether to preserve (``True``) or discard (``False``)
                    the basis element. By default (``None``), no filtering is applied.
            initialize (bool, optional): initialize the weights of the model. Default: ``True``

        """

        assert isinstance(in_type.gspace, GSpace3D)
        assert isinstance(out_type.gspace, GSpace3D)
        assert in_type.gspace == out_type.gspace
        assert isinstance(in_type.gspace.fibergroup, Icosahedral)

        assert samples in ['ico', 'dodeca', 'icosidodeca']
        self._samples = samples

        super(R3IcoConv, self).__init__(
            in_type,
            out_type,
            kernel_size,
            padding,
            stride,
            dilation,
            padding_mode,
            groups,
            bias,
            sigma,
            frequencies_cutoff=5., # to avoid performing any frequency cut-off, set a large upperbound
            rings=rings,
            recompute=recompute,
            basis_filter=basis_filter,
            initialize=initialize,
        )

    def _build_kernel_basis(self, in_repr: Representation, out_repr: Representation) -> KernelBasis:
        if self._samples == 'ico':
            return kernels_aliased_Ico_act_R3_icosahedron(in_repr, out_repr, sigma=self._sigma, radii=self._rings)
        if self._samples == 'dodeca':
            return kernels_aliased_Ico_act_R3_dodecahedron(in_repr, out_repr, sigma=self._sigma, radii=self._rings)
        if self._samples == 'icosidodeca':
            return kernels_aliased_Ico_act_R3_icosidodecahedron(in_repr, out_repr, sigma=self._sigma, radii=self._rings)
        else:
            raise ValueError






from collections import defaultdict


from escnn.gspaces import *
from escnn.nn import FieldType
from escnn.nn import GeometricTensor

from ..equivariant_module import EquivariantModule

import torch
from torch.nn import Parameter
from typing import List, Tuple, Any, Union
import numpy as np

from abc import ABC, abstractmethod

__all__ = [
    "_IIDBatchNorm",
    "IIDBatchNorm1d",
    "IIDBatchNorm2d",
    "IIDBatchNorm3d",
]


class _IIDBatchNorm(EquivariantModule, ABC):
    
    def __init__(self,
                 in_type: FieldType,
                 eps: float = 1e-05,
                 momentum: float = 0.1,
                 affine: bool = True,
                 track_running_stats: bool = True,
                 ):
    
        assert isinstance(in_type.gspace, GSpace)
        
        super(_IIDBatchNorm, self).__init__()

        self.space = in_type.gspace
        self.in_type = in_type
        self.out_type = in_type
        
        self.affine = affine
        self.track_running_stats = track_running_stats

        self._nfields = None
        
        # group fields by their type and
        #   - check if fields of the same type are contiguous
        #   - retrieve the indices of the fields

        # number of fields of each type
        self._nfields = defaultdict(int)
        
        # indices of the channels corresponding to fields belonging to each group
        _indices = defaultdict(lambda: [])
        
        # whether each group of fields is contiguous or not
        self._contiguous = {}
        
        ntrivials = 0
        position = 0
        last_field = None
        for i, r in enumerate(self.in_type.representations):
            
            for irr in r.irreps:
                if self.in_type.fibergroup.irrep(*irr).is_trivial():
                    ntrivials += 1
            
            if r.name != last_field:
                if not r.name in self._contiguous:
                    self._contiguous[r.name] = True
                else:
                    self._contiguous[r.name] = False

            last_field = r.name
            _indices[r.name] += list(range(position, position + r.size))
            self._nfields[r.name] += 1
            position += r.size
        
        for name, contiguous in self._contiguous.items():
            
            if contiguous:
                # for contiguous fields, only the first and last indices are kept
                _indices[name] = [min(_indices[name]), max(_indices[name])+1]
                setattr(self, f"{name}_indices", _indices[name])
            else:
                # otherwise, transform the list of indices into a tensor
                _indices[name] = torch.LongTensor(_indices[name])
                
                # register the indices tensors as parameters of this module
                self.register_buffer(f"{name}_indices", _indices[name])
        
        # store the size of each field type
        self._sizes = []
        
        # store for each field type the indices of the trivial irreps in it
        self._trivial_idxs = {}

        # store for each field type the sizes and the indices of all its irreps, grouped by their size
        self._irreps_sizes = {}
        
        self._has_trivial = {}

        # for each different representation in the input type
        for r in self.in_type._unique_representations:
            p = 0
            trivials = []
            
            # mask containing the location of the trivial irreps in the irrep decomposition of the representation
            S = np.zeros((r.size, r.size))
            
            # find all trivial irreps occurring in the representation
            for i, irr in enumerate(r.irreps):
                irr = self.in_type.fibergroup.irrep(*irr)
                if irr.is_trivial():
                    trivials.append(p)
                    S[p, p] = 1.
                
                p += irr.size
            
            name = r.name
            self._sizes.append((name, r.size))
            
            self._has_trivial[name] = len(trivials) > 0
            
            if self._has_trivial[name]:
                # averaging matrix which computes the expectation of a input vector, i.e. projects it in the trivial
                # subspace by masking out all non-trivial irreps
                P = r.change_of_basis @ S @ r.change_of_basis_inv
                self.register_buffer(f'{name}_avg', torch.tensor(P, dtype=torch.float))
            
                Q = torch.tensor(r.change_of_basis, dtype=torch.float)[:, trivials]
                self.register_buffer(f'{name}_change_of_basis', Q)
                
                running_mean = torch.zeros((self._nfields[r.name], r.size), dtype=torch.float)
                self.register_buffer(f'{name}_running_mean', running_mean)

            # assume all dimensions have same variance, i.e. the covariance matrix is a scalar multiple of the identity
            running_var = torch.ones((self._nfields[r.name], 1), dtype=torch.float)
            self.register_buffer(f'{name}_running_var', running_var)
            
            if self.affine:
                # scale all dimensions of the same field by the same weight
                weight = Parameter(torch.ones((self._nfields[r.name], 1)), requires_grad=True)
                self.register_parameter(f'{name}_weight', weight)
                if self._has_trivial[name]:
                    # the bias is applied only to the trivial channels
                    bias = Parameter(torch.zeros((self._nfields[r.name], len(trivials))), requires_grad=True)
                    self.register_parameter(f'{name}_bias', bias)
            
        self.register_buffer('num_batches_tracked', torch.tensor(0, dtype=torch.long))
        
        self.eps = eps
        self.momentum = momentum

    def reset_running_stats(self):
        for name, size in self._sizes:
            if hasattr(self, f"{name}_running_var"):
                running_var = getattr(self, f"{name}_running_var")
                running_var.fill_(1)
                
            if hasattr(self, f"{name}_running_mean"):
                running_mean = getattr(self, f"{name}_running_mean")
                running_mean.fill_(0)
                
        self.num_batches_tracked.zero_()

    def reset_parameters(self):
        self.reset_running_stats()
        if self.affine:
            for name, size in self._sizes:
                weight = getattr(self, f"{name}_weight")
                weight.data.fill_(1)
                if hasattr(self, f"{name}_bias"):
                    bias = getattr(self, f"{name}_bias")
                    bias.data.fill_(0)
    
    def _estimate_stats(self, slice, name: str):

        agg_axes = (0,) + tuple(range(3, len(slice.shape)))

        if self._has_trivial[name]:
            P = getattr(self, f'{name}_avg')

            # compute the mean
            means = torch.einsum(
                'ij,bcj...->bci...',
                P,
                slice.mean(dim=agg_axes, keepdim=True).detach()
            )
            centered = slice - means
            means = means.reshape(means.shape[1], means.shape[2])
        else:
            means = None
            centered = slice
    
        # Center the data and compute the variance
        # N.B.: we implicitly assume the dimensions to be iid,
        # i.e. the covariance matrix is a scalar multiple of the identity
        vars = centered.var(dim=agg_axes, unbiased=True, keepdim=False).mean(dim=1, keepdim=True).detach()

        return means, vars
    
    def _get_running_stats(self, name: str):
        vars = getattr(self, f"{name}_running_var")
        if self._has_trivial[name]:
            means = getattr(self, f"{name}_running_mean")
        else:
            means = None
            
        return means, vars

    def forward(self, input: GeometricTensor) -> GeometricTensor:
        r"""
        Apply norm non-linearities to the input feature map
        
        Args:
            input (GeometricTensor): the input feature map

        Returns:
            the resulting feature map
            
        """
        
        assert input.type == self.in_type

        exponential_average_factor = 0.0

        if self.training:
            self.num_batches_tracked += 1
            if self.momentum is None:  # use cumulative moving average
                exponential_average_factor = 1.0 / self.num_batches_tracked.item()
            else:  # use exponential moving average
                exponential_average_factor = self.momentum

        coords = input.coords
        input = input.tensor
        b, c = input.shape[:2]
        spatial_dims = input.shape[2:]

        output = torch.empty_like(input)
        
        # iterate through all field types
        for name, size in self._sizes:
            indices = getattr(self, f"{name}_indices")
            
            if self._contiguous[name]:
                slice = input[:, indices[0]:indices[1], ...]
            else:
                slice = input[:, indices, ...]

            slice = slice.view(b, -1, size, *spatial_dims)
            
            if not self.track_running_stats:
                means, vars = self._estimate_stats(slice, name)
                
            elif self.training:
                means, vars = self._estimate_stats(slice, name)
                running_mean, running_var = self._get_running_stats(name)

                running_var *= 1 - exponential_average_factor
                running_var += exponential_average_factor * vars
                assert torch.allclose(running_var, getattr(self, f"{name}_running_var"))

                if self._has_trivial[name]:
                    running_mean *= 1 - exponential_average_factor
                    running_mean += exponential_average_factor * means
                    assert torch.allclose(running_mean, getattr(self, f"{name}_running_mean"))
                
            else:
                means, vars = self._get_running_stats(name)

            if self._has_trivial[name]:
                # center data by subtracting the mean
                slice = slice - means.view(1, means.shape[0], means.shape[1], *(1,)*len(spatial_dims))

            # normalize dividing by the std and multiply by the new scale
            if self.affine:
                weight = getattr(self, f"{name}_weight")
            else:
                weight = 1.

            # compute the scalar multipliers needed
            scales = weight / (vars + self.eps).sqrt()
            # scale features
            slice = slice * scales.view(1, scales.shape[0], scales.shape[1], *(1,)*len(spatial_dims))
            
            # shift the features with the learnable bias
            if self.affine and self._has_trivial[name]:
                bias = getattr(self, f"{name}_bias")
                Q = getattr(self, f'{name}_change_of_basis')
                slice = slice + torch.einsum(
                    'ij,cj->ci',
                    Q,
                    bias
                ).view(1, bias.shape[0], Q.shape[0], *(1,)*len(spatial_dims))

            if not self._contiguous[name]:
                output[:, indices, ...] = slice.view(b, -1, *spatial_dims)
            else:
                output[:, indices[0]:indices[1], ...] = slice.view(b, -1, *spatial_dims)

        # wrap the result in a GeometricTensor
        return GeometricTensor(output, self.out_type, coords)

    def evaluate_output_shape(self, input_shape: Tuple[int, ...]) -> Tuple[int, ...]:
        assert len(input_shape) > 1
        assert input_shape[1] == self.in_type.size
    
        return (input_shape[0], self.out_type.size, *input_shape[2:])

    def check_equivariance(self, atol: float = 1e-6, rtol: float = 1e-5) -> List[Tuple[Any, float]]:
        # return super(NormBatchNorm, self).check_equivariance(atol=atol, rtol=rtol)
        pass

    def __repr__(self):
        extra_lines = []
        extra_repr = self.extra_repr()
        if extra_repr:
            extra_lines = extra_repr.split('\n')
    
        main_str = self._get_name() + '('
        if len(extra_lines) == 1:
            main_str += extra_lines[0]
        else:
            main_str += '\n  ' + '\n  '.join(extra_lines) + '\n'
    
        main_str += ')'
        return main_str

    def extra_repr(self):
        return '{in_type}, eps={eps}, momentum={momentum}, affine={affine}, track_running_stats={track_running_stats}' \
            .format(**self.__dict__)

    @abstractmethod
    def export(self):
        pass

    @abstractmethod
    def _check_input_shape(self, shape: Tuple[int, ...]):
        pass


class IIDBatchNorm1d(_IIDBatchNorm):
    r"""

    Batch normalization for generic representations for 1D or 0D data (i.e. 3D or 2D inputs).

    This batch normalization assumes that all dimensions within the same field have the same variance, i.e. that
    the covariance matrix of each field in `in_type` is a scalar multiple of the identity.
    Moreover, the mean is only computed over the trivial irreps occourring in the input representations (the input
    representation does not need to be decomposed into a direct sum of irreps since this module can deal with the
    change of basis).

    Similarly, if `affine = True`, a single scale is learnt per input field and the bias is applied only to the
    trivial irreps.

    This assumption is equivalent to the usual Batch Normalization in a Group Convolution NN (GCNN), where
    statistics are shared over the group dimension.
    See Chapter 4.2 at `https://gabri95.github.io/Thesis/thesis.pdf <https://gabri95.github.io/Thesis/thesis.pdf>`_ .

    Args:
        in_type (FieldType): the input field type
        eps (float, optional): a value added to the denominator for numerical stability. Default: ``1e-5``
        momentum (float, optional): the value used for the ``running_mean`` and ``running_var`` computation.
                Can be set to ``None`` for cumulative moving average (i.e. simple average). Default: ``0.1``
        affine (bool, optional): if ``True``, this module has learnable affine parameters. Default: ``True``
        track_running_stats (bool, optional): when set to ``True``, the module tracks the running mean and variance;
                                              when set to ``False``, it does not track such statistics but uses
                                              batch statistics in both training and eval modes.
                                              Default: ``True``

    """
    
    def _check_input_shape(self, shape: Tuple[int, ...]):
        if len(shape) != 2 or len(shape) != 3:
            raise ValueError('Error, expected a 2D or 3D tensor but a {} one was found'.format(len(shape)))
    
    def export(self):
        r"""
        Export this module to a normal PyTorch :class:`torch.nn.BatchNorm2d` module and set to "eval" mode.

        """
        
        if not self.track_running_stats:
            raise ValueError('''
                Equivariant Batch Normalization can not be converted into conventional batch normalization when
                "track_running_stats" is False because the statistics contained in a single batch are generally
                not symmetric
            ''')
        
        self.eval()
        
        batchnorm = torch.nn.BatchNorm1d(
            self.in_type.size,
            self.eps,
            self.momentum,
            affine=self.affine,
            track_running_stats=self.track_running_stats
        )
        
        batchnorm.num_batches_tracked.data = self.num_batches_tracked.data
        
        for name, size in self._sizes:
            contiguous = self._contiguous[name]
            if not contiguous:
                raise NotImplementedError(
                    '''Non-contiguous indices not supported yet when converting
                    inner-batch normalization into conventional BatchNorm2d'''
                )
            
            start, end = getattr(self, '{}_indices'.format(name))
            
            running_mean, running_var = self._get_running_stats(name)
            
            n = self._nfields[name]
            
            batchnorm.running_var.data[start:end] = running_var.data.view(n, 1).expand(n, size).reshape(-1)
            if self.affine:
                weight = getattr(self, '{}_weight'.format(name))
                batchnorm.weight.data[start:end] = weight.data.view(n, 1).expand(n, size).reshape(-1)
            
            if self._has_trivial[name]:
                batchnorm.running_mean.data[start:end] = running_mean.data.view(n, size).reshape(-1)
                if self.affine:
                    bias = getattr(self, '{}_bias'.format(name))
                    Q = getattr(self, '{}_change_of_basis'.format(name))
                    bias = torch.einsum(
                        'ij,cj->ci',
                        Q,
                        bias
                    )
                    batchnorm.bias.data[start:end] = bias.data.view(n, size).reshape(-1)
            else:
                batchnorm.running_mean.data[start:end] = 0.
                if self.affine:
                    batchnorm.bias.data[start:end] = 0.
        
        batchnorm.eval()
        
        return batchnorm


class IIDBatchNorm2d(_IIDBatchNorm):
    r"""

    Batch normalization for generic representations for 2D data (i.e. 4D inputs).
    
    This batch normalization assumes that all dimensions within the same field have the same variance, i.e. that
    the covariance matrix of each field in `in_type` is a scalar multiple of the identity.
    Moreover, the mean is only computed over the trivial irreps occourring in the input representations (the input
    representation does not need to be decomposed into a direct sum of irreps since this module can deal with the
    change of basis).
    
    Similarly, if `affine = True`, a single scale is learnt per input field and the bias is applied only to the
    trivial irreps.
    
    This assumption is equivalent to the usual Batch Normalization in a Group Convolution NN (GCNN), where
    statistics are shared over the group dimension.
    See Chapter 4.2 at `https://gabri95.github.io/Thesis/thesis.pdf <https://gabri95.github.io/Thesis/thesis.pdf>`_ .

    Args:
        in_type (FieldType): the input field type
        eps (float, optional): a value added to the denominator for numerical stability. Default: ``1e-5``
        momentum (float, optional): the value used for the ``running_mean`` and ``running_var`` computation.
                Can be set to ``None`` for cumulative moving average (i.e. simple average). Default: ``0.1``
        affine (bool, optional): if ``True``, this module has learnable affine parameters. Default: ``True``
        track_running_stats (bool, optional): when set to ``True``, the module tracks the running mean and variance;
                                              when set to ``False``, it does not track such statistics but uses
                                              batch statistics in both training and eval modes.
                                              Default: ``True``

    """

    def _check_input_shape(self, shape: Tuple[int, ...]):
        if len(shape) != 4:
            raise ValueError('Error, expected a 4D tensor but a {} one was found'.format(len(shape)))

    def export(self):
        r"""
        Export this module to a normal PyTorch :class:`torch.nn.BatchNorm2d` module and set to "eval" mode.

        """

        if not self.track_running_stats:
            raise ValueError('''
                Equivariant Batch Normalization can not be converted into conventional batch normalization when
                "track_running_stats" is False because the statistics contained in a single batch are generally
                not symmetric
            ''')

        self.eval()

        batchnorm = torch.nn.BatchNorm2d(
            self.in_type.size,
            self.eps,
            self.momentum,
            affine=self.affine,
            track_running_stats=self.track_running_stats
        )

        batchnorm.num_batches_tracked.data = self.num_batches_tracked.data

        for name, size in self._sizes:
            contiguous = self._contiguous[name]
            if not contiguous:
                raise NotImplementedError(
                    '''Non-contiguous indices not supported yet when converting
                    inner-batch normalization into conventional BatchNorm2d'''
                )

            start, end = getattr(self, '{}_indices'.format(name))

            running_mean, running_var = self._get_running_stats(name)

            n = self._nfields[name]

            batchnorm.running_var.data[start:end] = running_var.data.view(n, 1).expand(n, size).reshape(-1)
            if self.affine:
                weight = getattr(self, '{}_weight'.format(name))
                batchnorm.weight.data[start:end] = weight.data.view(n, 1).expand(n, size).reshape(-1)

            if self._has_trivial[name]:
                batchnorm.running_mean.data[start:end] = running_mean.data.view(n, size).reshape(-1)
                if self.affine:
                    bias = getattr(self, '{}_bias'.format(name))
                    Q = getattr(self, '{}_change_of_basis'.format(name))
                    bias = torch.einsum(
                        'ij,cj->ci',
                        Q,
                        bias
                    )
                    batchnorm.bias.data[start:end] = bias.data.view(n, size).reshape(-1)
            else:
                batchnorm.running_mean.data[start:end] = 0.
                if self.affine:
                    batchnorm.bias.data[start:end] = 0.

        batchnorm.eval()

        return batchnorm


class IIDBatchNorm3d(_IIDBatchNorm):
    r"""

    Batch normalization for generic representations for 3D data (i.e. 5D inputs).
    
    This batch normalization assumes that all dimensions within the same field have the same variance, i.e. that
    the covariance matrix of each field in `in_type` is a scalar multiple of the identity.
    Moreover, the mean is only computed over the trivial irreps occourring in the input representations (the input
    representation does not need to be decomposed into a direct sum of irreps since this module can deal with the
    change of basis).
    
    Similarly, if `affine = True`, a single scale is learnt per input field and the bias is applied only to the
    trivial irreps.
    
    This assumption is equivalent to the usual Batch Normalization in a Group Convolution NN (GCNN), where
    statistics are shared over the group dimension.
    See Chapter 4.2 at `https://gabri95.github.io/Thesis/thesis.pdf <https://gabri95.github.io/Thesis/thesis.pdf>`_ .

    Args:
        in_type (FieldType): the input field type
        eps (float, optional): a value added to the denominator for numerical stability. Default: ``1e-5``
        momentum (float, optional): the value used for the ``running_mean`` and ``running_var`` computation.
                Can be set to ``None`` for cumulative moving average (i.e. simple average). Default: ``0.1``
        affine (bool, optional): if ``True``, this module has learnable affine parameters. Default: ``True``
        track_running_stats (bool, optional): when set to ``True``, the module tracks the running mean and variance;
                                              when set to ``False``, it does not track such statistics but uses
                                              batch statistics in both training and eval modes.
                                              Default: ``True``

    """

    def _check_input_shape(self, shape: Tuple[int, ...]):
        if len(shape) != 5:
            raise ValueError('Error, expected a 5D tensor but a {} one was found'.format(len(shape)))

    def export(self):
        r"""
        Export this module to a normal PyTorch :class:`torch.nn.BatchNorm2d` module and set to "eval" mode.

        """

        if not self.track_running_stats:
            raise ValueError('''
                Equivariant Batch Normalization can not be converted into conventional batch normalization when
                "track_running_stats" is False because the statistics contained in a single batch are generally
                not symmetric
            ''')

        self.eval()

        batchnorm = torch.nn.BatchNorm3d(
            self.in_type.size,
            self.eps,
            self.momentum,
            affine=self.affine,
            track_running_stats=self.track_running_stats
        )

        batchnorm.num_batches_tracked.data = self.num_batches_tracked.data

        for name, size in self._sizes:
            contiguous = self._contiguous[name]
            if not contiguous:
                raise NotImplementedError(
                    '''Non-contiguous indices not supported yet when converting
                    inner-batch normalization into conventional BatchNorm2d'''
                )

            start, end = getattr(self, '{}_indices'.format(name))

            running_mean, running_var = self._get_running_stats(name)

            n = self._nfields[name]

            batchnorm.running_var.data[start:end] = running_var.data.view(n, 1).expand(n, size).reshape(-1)
            if self.affine:
                weight = getattr(self, '{}_weight'.format(name))
                batchnorm.weight.data[start:end] = weight.data.view(n, 1).expand(n, size).reshape(-1)

            if self._has_trivial[name]:
                batchnorm.running_mean.data[start:end] = running_mean.data.view(n, size).reshape(-1)
                if self.affine:
                    bias = getattr(self, '{}_bias'.format(name))
                    Q = getattr(self, '{}_change_of_basis'.format(name))
                    bias = torch.einsum(
                        'ij,cj->ci',
                        Q,
                        bias
                    )
                    batchnorm.bias.data[start:end] = bias.data.view(n, size).reshape(-1)
            else:
                batchnorm.running_mean.data[start:end] = 0.
                if self.affine:
                    batchnorm.bias.data[start:end] = 0.

        batchnorm.eval()

        return batchnorm





from .field_type import FieldType
from .geometric_tensor import GeometricTensor, tensor_directsum

from .modules import *
from .modules import __all__ as modules_list

__all__ = [
    "FieldType",
    "GeometricTensor",
    "tensor_directsum",
    # Modules
] + modules_list + [
    # init
    "init",
]




from collections import defaultdict

from torch.nn import Parameter

from escnn.gspaces import *
from escnn.nn import FieldType
from escnn.nn import GeometricTensor

from ..equivariant_module import EquivariantModule

import torch

from typing import List, Tuple, Any

import numpy as np

__all__ = ["NormNonLinearity"]


class NormNonLinearity(EquivariantModule):
    
    def __init__(self, in_type: FieldType, function: str = 'n_relu', bias: bool = True):
        r"""
        
        Norm non-linearities.
        This module applies a bias and an activation function over the norm of each field.
        
        The input representation of the fields is preserved by this operation.
        
        .. note ::
            If 'squash' non-linearity (`function`) is chosen, no bias is allowed
        
        Args:
            in_type (FieldType): the input field type
            function (str, optional): the identifier of the non-linearity. It is used to specify which function to
                                      apply. By default (``'n_relu'``), ReLU is used.
            bias (bool, optional): add bias to norm of fields before computing the non-linearity. Default: ``True``

        """

        assert isinstance(in_type.gspace, GSpace)
        
        super(NormNonLinearity, self).__init__()

        for r in in_type.representations:
            assert 'norm' in r.supported_nonlinearities, \
                'Error! Representation "{}" does not support "norm" non-linearity'.format(r.name)

        self.space = in_type.gspace
        self.in_type = in_type
        self.out_type = in_type
        
        self._nfields = None
        self.log_bias = None

        if function == 'n_relu':
            self._function = torch.relu
        elif function == 'n_sigmoid':
            self._function = torch.sigmoid
        elif function == 'n_softplus':
            self._function = torch.nn.functional.softplus
        elif function == "squash":
            self._function = lambda t: t / (1.0 + t)
            assert bias is False, 'Error! When using squash non-linearity, norm bias is not allowed'
        else:
            raise ValueError('Function "{}" not recognized!'.format(function))
        
        # group fields by their size and
        #   - check if fields of the same size are contiguous
        #   - retrieve the indices of the fields

        # number of fields of each size
        self._nfields = defaultdict(int)
        
        # indices of the channales corresponding to fields belonging to each group
        _indices = defaultdict(lambda: [])
        
        # whether each group of fields is contiguous or not
        self._contiguous = {}
        
        position = 0
        last_size = None
        for i, r in enumerate(self.in_type.representations):
            
            if r.size != last_size:
                if not r.size in self._contiguous:
                    self._contiguous[r.size] = True
                else:
                    self._contiguous[r.size] = False
            last_size = r.size
            
            _indices[r.size] += list(range(position, position + r.size))
            self._nfields[r.size] += 1
            position += r.size
        
        for s, contiguous in self._contiguous.items():
            if contiguous:
                # for contiguous fields, only the first and last indices are kept
                _indices[s] = torch.LongTensor([min(_indices[s]), max(_indices[s])+1])
            else:
                # otherwise, transform the list of indices into a tensor
                _indices[s] = torch.LongTensor(_indices[s])
                
            # register the indices tensors as parameters of this module
            self.register_buffer('indices_{}'.format(s), _indices[s])
        
        if bias:
            # build a bias for each field
            self.log_bias = Parameter(torch.zeros(1, len(self.in_type), 1, 1, dtype=torch.float), requires_grad=True)
        else:
            self.log_bias = None
    
        # build a sorted list of the fields groups, such that every time they are iterated through in the same order
        self._order = sorted(self._contiguous.keys())
        
        self.eps = Parameter(torch.tensor(1e-10), requires_grad=False)
    
    def forward(self, input: GeometricTensor) -> GeometricTensor:
        r"""
        Apply norm non-linearities to the input feature map
        
        Args:
            input (GeometricTensor): the input feature map

        Returns:
            the resulting feature map
            
        """
        
        assert input.type == self.in_type

        coords = input.coords
        input = input.tensor
        
        # scalar multipliers needed to turn the old norms into the newly computed ones
        multipliers = torch.empty_like(input)

        b, c = input.shape[:2]
        spatial_dims = input.shape[2:]

        next_bias = 0
        
        if self.log_bias is not None:
            # build the bias
            # biases = torch.nn.functional.elu(self.log_bias)
            biases = torch.exp(self.log_bias)
            # biases = torch.nn.functional.elu(self.log_bias) + 1
        else:
            biases = None
        
        # iterate through all field sizes
        for s in self._order:
            
            # retrieve the corresponding fiber indices
            indices = getattr(self, f"indices_{s}")
            
            if self._contiguous[s]:
                # if the fields were contiguous, we can use slicing
                # retrieve the fields
                fm = input[:, indices[0]:indices[1], ...]
            else:
                # otherwise we have to use indexing
                # retrieve the fields
                fm = input[:, indices, ...]

            # compute the norm of each field
            norms = fm.view(b, -1, s, *spatial_dims).norm(dim=2, keepdim=True)
            
            # compute the new norms
            if biases is not None:
                # retrieve the bias elements corresponding to the current fields
                bias = biases[:, next_bias:next_bias + self._nfields[s], ...].view(1, -1, 1, *[1]*len(spatial_dims))
                new_norms = self._function(norms - bias)
            else:
                new_norms = self._function(norms)

            # compute the scalar multipliers needed to turn the old norms into the newly computed ones
            # m = torch.zeros_like(new_norms)
            # in order to avoid division by 0
            # mask = norms > 0.
            # m[mask] = new_norms[mask] / norms[mask]
            
            m = new_norms / torch.max(norms, self.eps)
            m[norms <= self.eps] = 0.

            if self._contiguous[s]:
                # expand the multipliers tensor to all channels for each field
                multipliers[:, indices[0]:indices[1], ...] = m.expand(b, -1, s, *spatial_dims).reshape(b, -1, *spatial_dims)
            
            else:
                # expand the multipliers tensor to all channels for each field
                multipliers[:, indices, ...] = m.expand(b, -1, s, *spatial_dims).reshape(b, -1, *spatial_dims)
            
            # shift the position on the bias tensor
            next_bias += self._nfields[s]
        
        # multiply the input by the multipliers computed and wrap the result in a GeometricTensor
        return GeometricTensor(input * multipliers, self.out_type, coords)

    def evaluate_output_shape(self, input_shape: Tuple[int, ...]) -> Tuple[int, ...]:

        assert len(input_shape) >= 2
        assert input_shape[1] == self.in_type.size

        b, c = input_shape[:2]
        spatial_shape = input_shape[2:]

        return (b, self.out_type.size, *spatial_shape)

    def check_equivariance(self, atol: float = 1e-6, rtol: float = 1e-5) -> List[Tuple[Any, float]]:
    
        c = self.in_type.size
    
        x = torch.randn(3, c, 10, 10)
    
        x = GeometricTensor(x, self.in_type)
    
        errors = []
    
        for el in self.space.testing_elements:
            out1 = self(x).transform_fibers(el)
            out2 = self(x.transform_fibers(el))
        
            errs = (out1.tensor - out2.tensor).detach().numpy()
            errs = np.abs(errs).reshape(-1)
            print(el, errs.max(), errs.mean(), errs.var())
        
            assert torch.allclose(out1.tensor, out2.tensor, atol=atol, rtol=rtol), \
                'The error found during equivariance check with element "{}" is too high: max = {}, mean = {} var ={}' \
                    .format(el, errs.max(), errs.mean(), errs.var())
        
            errors.append((el, errs.mean()))
    
        return errors
